@@ -1,9 +1,9 @@
-# Copyright 2001-2005 The Apache Software Foundation or its licensors, as
-# applicable.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# Licensed to the Apache Software Foundation (ASF) under one or more
+# contributor license agreements.  See the NOTICE file distributed with
+# this work for additional information regarding copyright ownership.
+# The ASF licenses this file to You under the Apache License, Version 2.0
+# (the "License"); you may not use this file except in compliance with
+# the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
@@ -21,77 +21,112 @@ use warnings FATAL => 'all';
 use Exporter ();
 use Config;
 use Apache::TestConfig ();
+use Test qw/ok skip/;
+
+BEGIN {
+    # Apache::Test loads a bunch of mp2 stuff while getting itself
+    # together.  because we need to choose one of mp1 or mp2 to load
+    # check first (and we choose mp2) $mod_perl::VERSION == 2.0
+    # just because someone loaded Apache::Test.  This Is Bad.  so,
+    # let's try to correct for that here by removing mod_perl from
+    # %INC after the above use() statements settle in.  nobody
+    # should be relying on us loading up mod_perl.pm anyway...
+
+    delete $INC{'mod_perl.pm'};
+}
 
 use vars qw(@ISA @EXPORT %EXPORT_TAGS $VERSION %SubTests @SkipReasons);
 
-$VERSION = '1.27';
+$VERSION = '1.39';
 
 my @need = qw(need_lwp need_http11 need_cgi need_access need_auth
               need_module need_apache need_min_apache_version
               need_apache_version need_perl need_min_perl_version
-              need_min_module_version need_threads need_apache_mpm
-              need_php need_php4 need_ssl);
+              need_min_module_version need_threads need_fork need_apache_mpm
+              need_php need_php4 need_ssl need_imagemap need_cache_disk);
 
 my @have = map { (my $need = $_) =~ s/need/have/; $need } @need;
 
 @ISA = qw(Exporter);
-@EXPORT = (qw(ok skip sok plan skip_reason under_construction need),
+@EXPORT = (qw(sok plan skip_reason under_construction need),
            @need, @have);
 
-# everything but ok(), skip(), and plan() - Test::More provides these
-my @test_more_exports = grep { ! /^(ok|skip|plan)$/ } @EXPORT;
-
-%EXPORT_TAGS = (withtestmore => \@test_more_exports);
-
 %SubTests = ();
 @SkipReasons = ();
 
-if (my $subtests = $ENV{HTTPD_TEST_SUBTESTS}) {
-    %SubTests = map { $_, 1 } split /\s+/, $subtests;
+sub cp {
+    my @l;
+    for( my $i=1; (@l=caller $i)[0] eq __PACKAGE__; $i++ ) {};
+    return wantarray ? @l : $l[0];
 }
 
 my $Config;
-my $real_plan;
-my @testmore;
-
+my %wtm;
 sub import {
-    my $class = shift;
+    my $class=$_[0];
+    my $wtm=0;
+    my @base_exp;
+    my @exp;
+    my %my_exports;
+    undef @my_exports{@EXPORT};
 
-    # once Test::More always Test::More until plan() is called
-    if (($_[0] and $_[0] =~ m/^-withtestmore/) || @testmore) {
-        # special hoops for Test::More support
+    my ($caller,$f,$l)=cp;
 
-        $real_plan = eval { 
-
-            require Test::More; 
-
-            no warnings qw(numeric);
-            Test::Builder->VERSION('0.18_01');
-
-            # required for Test::More::import() and Apache::Test::plan()
-            # if we don't do this, Test::More exports plan() anyway
-            # and we get collisions.  go figure.
-            @testmore = (import => [qw(!plan)]);
-
-            Test::More->import(@testmore);
-
-            \&Test::More::plan;
-        } or die "-withtestmore error: $@";
-
-        # clean up arguments to export_to_level
-        shift;
-        @EXPORT = (@test_more_exports, @Test::More::EXPORT);
+    for( my $i=1; $i<@_; $i++ ) {
+	if( $_[$i] eq '-withtestmore' ) {
+	    $wtm=1;
+	}
+	elsif( $_[$i] eq ':DEFAULT' ) {
+	    push @exp, $_[$i];
+	    push @base_exp, $_[$i];
+	}
+	elsif( $_[$i] eq '!:DEFAULT' ) {
+	    push @exp, $_[$i];
+	    push @base_exp, $_[$i];
+	}
+	elsif( $_[$i]=~m@^[:/!]@ ) {
+	    warn("Ignoring import spec $_[$i] ".
+		 "at $f line $l\n")
+	}
+	elsif( exists $my_exports{$_[$i]} ) {
+	    push @exp, $_[$i];
+	}
+	else {
+	    push @base_exp, $_[$i];
+	}
     }
-    else {
-        # the default - Test.pm support
-
-        require Test;
-        Test->import(qw(ok skip));
-        @testmore = ();               # reset, just in case.
-        $real_plan = \&Test::plan;
+    if (!@exp and @base_exp) {
+	@exp=('!:DEFAULT');
+    }
+    elsif (@exp and !@base_exp) {
+	@base_exp=('!:DEFAULT');
     }
 
-    $class->export_to_level(1, undef, @_ ? @_ : @EXPORT);
+    $wtm{$caller}=[$wtm,$f,$l] unless exists $wtm{$caller};
+
+    warn("Ignoring -withtestmore due to a previous call ".
+	 "($wtm{$caller}->[1]:$wtm{$caller}->[2]) without it ".
+	 "at $f line $l\n")
+	if $wtm{$caller}->[0]==0 and $wtm==1;
+
+    $class->export_to_level(1, $class, @exp);
+
+    push @base_exp, '!plan';
+    if( $wtm{$caller}->[0] ) {	# -withtestmore
+	eval <<"EVAL"
+package $caller;
+#line $l $f
+use Test::More import=>\\\@base_exp;
+EVAL
+    }
+    else {			# -withouttestmore
+	eval <<"EVAL";
+package $caller;
+#line $l $f
+use Test \@base_exp;
+EVAL
+    }
+    die $@ if $@;
 }
 
 sub config {
@@ -113,26 +148,45 @@ sub sok (&;$) {
     my $sub = shift;
     my $nok = shift || 1; #allow sok to have 'ok' within
 
-    if (%SubTests and not $SubTests{ $Test::ntest }) {
-        for my $n (1..$nok) {
-            skip("skipping this subtest", 0);
-        }
-        return;
-    }
+    my ($caller,$f,$l)=cp;
 
-    my($package, $filename, $line) = caller;
+    if (exists $wtm{$caller} and $wtm{$caller}->[0]==1) { # -withtestmore
+	require Test::Builder;
+	my $tb=Test::Builder->new;
 
-    # trick ok() into reporting the caller filename/line when a
-    # sub-test fails in sok()
-    return eval <<EOE;
-#line $line $filename
-    ok(\$sub->());
+	if (%SubTests and not $SubTests{ 1+$tb->current_test }) {
+	    $tb->skip("skipping this subtest") for (1..$nok);
+	    return;
+	}
+
+	# trick ok() into reporting the caller filename/line when a
+	# sub-test fails in sok()
+	return eval <<EOE;
+#line $l $f
+    Test::More::ok(\$sub->());
 EOE
+    }
+    else {
+	if (%SubTests and not $SubTests{ $Test::ntest }) {
+	    skip("skipping this subtest", 0) for (1..$nok);
+	    return;
+	}
+
+	# trick ok() into reporting the caller filename/line when a
+	# sub-test fails in sok()
+	return eval <<EOE;
+#line $l $f
+    Test::ok(\$sub->());
+EOE
+    }
 }
 
 #so Perl's Test.pm can be run inside mod_perl
 sub test_pm_refresh {
-    if (@testmore) {
+    my ($caller,$f,$l)=cp;
+
+    if (exists $wtm{$caller} and $wtm{$caller}->[0]==1) { # -withtestmore
+	require Test::Builder;
         my $builder = Test::Builder->new;
 
         $builder->reset;
@@ -140,18 +194,30 @@ sub test_pm_refresh {
         $builder->output(\*STDOUT);
         $builder->todo_output(\*STDOUT);
 
-        # this is STDOUT because Test::More seems to put 
+        # this is STDOUT because Test::More seems to put
         # most of the stuff we want on STDERR, so it ends
         # up in the error_log instead of where the user can
         # see it.   consider leaving it alone based on
         # later user reports.
         $builder->failure_output(\*STDOUT);
     }
-    else {
-        $Test::TESTOUT = \*STDOUT;
-        $Test::planned = 0;
-        $Test::ntest = 1;
-        %Test::todo = ();
+    else {                                                # -withouttestmore
+	unless (exists $wtm{$caller}) {
+	    warn "You forgot to 'use Apache::Test' in package $caller\n";
+	    $wtm{$caller}=[0,$f,$l];
+	}
+	if (defined &Test::_reset_globals) {
+	    Test::_reset_globals();
+	    # Test.pm uses $TESTOUT=*STDOUT{IO}. We cannot do that
+	    # due to the way SetupEnv works.
+	    $Test::TESTOUT = \*STDOUT;
+	}
+	else {
+	    $Test::TESTOUT = \*STDOUT;
+	    $Test::planned = 0;
+	    $Test::ntest = 1;
+	    %Test::todo = ();
+	}
     }
 }
 
@@ -174,34 +240,6 @@ sub init_test_pm {
     $r->content_type('text/plain');
 }
 
-sub need_http11 {
-    require Apache::TestRequest;
-    if (Apache::TestRequest::install_http11()) {
-        return 1;
-    }
-    else {
-        push @SkipReasons,
-           "LWP version 5.60+ required for HTTP/1.1 support";
-        return 0;
-    }
-}
-
-sub need_ssl {
-    my $vars = vars();
-    need_module([$vars->{ssl_module_name}, 'Net::SSL']);
-}
-
-sub need_lwp {
-    require Apache::TestRequest;
-    if (Apache::TestRequest::has_lwp()) {
-        return 1;
-    }
-    else {
-        push @SkipReasons, "libwww-perl is not installed";
-        return 0;
-    }
-}
-
 sub plan {
     init_test_pm(shift) if ref $_[0];
     test_pm_refresh();
@@ -242,12 +280,56 @@ sub plan {
     }
     @SkipReasons = (); # reset
 
-    $real_plan->(@_, @testmore);
+    my ($caller,$f,$l)=cp;
+
+    %SubTests=();
+    if (my $subtests=$ENV{HTTPD_TEST_SUBTESTS}) {
+	%SubTests=map { $_, 1 } split /\s+/, $subtests;
+    }
+
+    if (exists $wtm{$caller} and $wtm{$caller}->[0]==1) { # -withtestmore
+	Test::More::plan(@_);
+    }
+    else {                                                # -withouttestmore
+	unless (exists $wtm{$caller}) {
+	    warn "You forgot to 'use Apache::Test' in package $caller\n";
+	    $wtm{$caller}=[0,$f,$l];
+	}
+	Test::plan(@_);
+    }
 
     # add to Test.pm verbose output
     print "# Using Apache/Test.pm version $VERSION\n";
 }
 
+sub need_http11 {
+    require Apache::TestRequest;
+    if (Apache::TestRequest::install_http11()) {
+        return 1;
+    }
+    else {
+        push @SkipReasons,
+           "LWP version 5.60+ required for HTTP/1.1 support";
+        return 0;
+    }
+}
+
+sub need_ssl {
+    my $vars = vars();
+    need_module([$vars->{ssl_module_name}, 'Net::SSL']);
+}
+
+sub need_lwp {
+    require Apache::TestRequest;
+    if (Apache::TestRequest::has_lwp()) {
+        return 1;
+    }
+    else {
+        push @SkipReasons, "libwww-perl is not installed";
+        return 0;
+    }
+}
+
 sub need {
     my $need_all = 1;
     for my $cond (@_) {
@@ -336,23 +418,54 @@ sub need_min_module_version {
 }
 
 sub need_cgi {
-    need_module('cgi') || need_module('cgid');
+    return _need_multi(qw(cgi.c cgid.c));
 }
 
+sub need_cache_disk {
+    return _need_multi(qw(cache_disk.c disk_cache.c));
+}
+
+
 sub need_php {
-    need_module('php4') || need_module('php5') || need_module('sapi_apache2.c');
+    return _need_multi(qw(php4 php5 sapi_apache2.c));
 }
 
 sub need_php4 {
-    need_module('php4') || need_module('sapi_apache2.c');
+    return _need_multi(qw(php4 sapi_apache2.c));
 }
 
 sub need_access {
-    need_module('access') || need_module('authz_host');
+    return _need_multi(qw(access authz_host));
 }
 
 sub need_auth {
-    need_module('auth') || need_module('auth_basic');
+    return _need_multi(qw(auth auth_basic));
+}
+
+sub need_imagemap {
+    return need_module("imagemap") || need_module("imap");
+}
+
+sub _need_multi {
+
+    my @check = @_;
+
+    my $rc = 0;
+
+    {
+        local @SkipReasons;
+
+        foreach my $module (@check) {
+          $rc ||= need_module($module);
+        }
+    }
+
+    my $reason = join ' or ', @check;
+
+    push @SkipReasons, "cannot find one of $reason"
+        unless $rc;
+
+    return $rc;
 }
 
 sub need_apache {
@@ -481,6 +594,22 @@ sub need_threads {
     return $status;
 }
 
+sub need_fork {
+    my $have_fork = $Config{d_fork} ||
+                    $Config{d_pseudofork} ||
+                    (($^O eq 'MSWin32' || $^O eq 'NetWare') &&
+                     $Config{useithreads} &&
+                     $Config{ccflags} =~ /-DPERL_IMPLICIT_SYS/);
+
+     if (!$have_fork) {
+         push @SkipReasons, 'The fork function is unimplemented';
+         return 0;
+     }
+     else {
+         return 1;
+     }
+}
+
 sub under_construction {
     push @SkipReasons, "This test is under construction";
     return 0;
@@ -518,6 +647,8 @@ for my $func (@have) {
 
 package Apache::TestToString;
 
+Apache::Test->import('!:DEFAULT');
+
 sub TIEHANDLE {
     my $string = "";
     bless \$string;
@@ -598,7 +729,7 @@ But this won't hint the reason for skipping therefore it's better to
 use need():
 
   plan tests => 5,
-      need 'LWP', 
+      need 'LWP',
            { "not Win32" => sub { $^O eq 'MSWin32'} };
 
 see C<need()> for more info.
@@ -609,6 +740,12 @@ need_module() is called for each value in this array. The test is
 skipped if need_module() returns false (which happens when at least
 one C or Perl module from the list cannot be found).
 
+Watch out for case insensitive file systems or duplicate modules
+with the same name.  I.E.  If you mean mod_env.c
+   need_module('mod_env.c')
+Not
+   need_module('env')
+
 =item * a C<CODE> reference
 
 the tests will be skipped if the function returns a false value. For
@@ -660,7 +797,7 @@ with C<plan()> to decide whether a test will run, but C<have_apache()>
 within test logic to adjust expectations based on older or newer
 server versions.
 
-=over 
+=over
 
 =item need_http11
 
@@ -688,6 +825,13 @@ Require LWP support.
 
 Requires mod_cgi or mod_cgid to be installed.
 
+=item need_cache_disk
+
+  plan tests => 5, need_cache_disk
+
+Requires mod_cache_disk or mod_disk_cache to be installed.
+
+
 =item need_php
 
   plan tests => 5, need_php;
@@ -700,6 +844,12 @@ Requires a PHP module to be installed (version 4 or 5).
 
 Requires a PHP version 4 module to be installed.
 
+=item need_imagemap
+
+  plan tests => 5, need_imagemap;
+
+Requires a mod_imagemap or mod_imap be installed
+
 =item need_apache
 
   plan tests => 5, need_apache 2;
@@ -765,6 +915,10 @@ For example:
 
 requires Perl 5.8.1 or higher.
 
+=item need_fork
+
+Requires the perl built-in function C<fork> to be implemented.
+
 =item need_module
 
   plan tests => 5, need_module 'CGI';
@@ -812,7 +966,7 @@ Currently works only for perl modules.
 
   plan tests => 5,
       need 'LWP',
-           { "perl >= 5.8.0 and w/ithreads is required" => 
+           { "perl >= 5.8.0 and w/ithreads is required" =>
              ($Config{useperlio} && $] >= 5.008) },
            { "not Win32"                 => sub { $^O eq 'MSWin32' },
              "foo is disabled"           => \&is_foo_enabled,
@@ -920,11 +1074,11 @@ are returned.
 
 =head1 Test::More Integration
 
-There are a few caveats if you want to use I<Apache::Test> with 
+There are a few caveats if you want to use I<Apache::Test> with
 I<Test::More> instead of the default I<Test> backend.  The first is
 that I<Test::More> requires you to use its own C<plan()> function
 and not the one that ships with I<Apache::Test>.  I<Test::More> also
-defines C<ok()> and C<skip()> functions that are different, and 
+defines C<ok()> and C<skip()> functions that are different, and
 simply C<use>ing both modules in your test script will lead to redefined
 warnings for these subroutines.
 
@@ -940,7 +1094,7 @@ with I<Test::More>.
 
     ok ('yes', 'testing ok');  # Test::More::ok()
 
-Now, while this works fine for standard client-side tests 
+Now, while this works fine for standard client-side tests
 (such as C<t/basic.t>), the more advanced features of I<Apache::Test>
 require using I<Test::More> as the sole driver behind the scenes.
 
@@ -964,7 +1118,7 @@ C<-withtestmore> tells I<Apache::Test> to use I<Test::More>
 instead of I<Test.pm> behind the scenes.  Note that you are not
 required to C<use Test::More> yourself with the C<-withtestmore>
 option and that the C<use Test::More tests =E<gt> 1> syntax
-may have unexpected results.  
+may have unexpected results.
 
 Note that I<Test::More> version 0.49, available within the
 I<Test::Simple> 0.49 distribution on CPAN, or greater is required
