@@ -1,9 +1,9 @@
-# Copyright 2001-2005 The Apache Software Foundation or its licensors, as
-# applicable.
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# Licensed to the Apache Software Foundation (ASF) under one or more
+# contributor license agreements.  See the NOTICE file distributed with
+# this work for additional information regarding copyright ownership.
+# The ASF licenses this file to You under the Apache License, Version 2.0
+# (the "License"); you may not use this file except in compliance with
+# the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
@@ -23,7 +23,8 @@ use constant OSX     => $^O eq 'darwin';
 use constant CYGWIN  => $^O eq 'cygwin';
 use constant NETWARE => $^O eq 'NetWare';
 use constant SOLARIS => $^O eq 'solaris';
-use constant WINFU   => WIN32 || CYGWIN || NETWARE;
+use constant AIX     => $^O eq 'aix';
+use constant WINFU   => WIN32 || NETWARE;
 use constant COLOR   => ($ENV{APACHE_TEST_COLOR} && -t STDOUT) ? 1 : 0;
 
 use constant DEFAULT_PORT => 8529;
@@ -32,12 +33,11 @@ use constant IS_MOD_PERL_2       =>
     eval { require mod_perl2 } || 0;
 
 use constant IS_MOD_PERL_2_BUILD => IS_MOD_PERL_2 &&
-    require Apache2::Build && Apache2::Build::IS_MOD_PERL_BUILD();
+    eval { require Apache2::Build && Apache2::Build::IS_MOD_PERL_BUILD() };
 
 use constant IS_APACHE_TEST_BUILD =>
-    grep { -e "$_/lib/Apache/TestConfig.pm" } qw(Apache-Test . ..);
-
-use constant CUSTOM_CONFIG_FILE => 'Apache/TestConfigData.pm';
+    grep { -e "$_/lib/Apache/TestConfig.pm" }
+         qw(Apache-Test . .. ../Apache-Test);
 
 use lib ();
 use File::Copy ();
@@ -58,23 +58,12 @@ use Apache::TestRun ();
 
 use vars qw(%Usage);
 
-# variables stored in $Apache::TestConfigData::vars
-my @data_vars_must = qw(httpd apxs);
-my @data_vars_opt  = qw(user group port);
-# mapping from $Apache::TestConfigData::vars to $ENV settings
-my %vars_to_env = (
-    httpd => 'APACHE_TEST_HTTPD',
-    apxs  => 'APACHE_TEST_APXS',
-    user  => 'APACHE_TEST_USER',
-    group => 'APACHE_TEST_GROUP',
-    port  => 'APACHE_TEST_PORT',
-);
-
 %Usage = (
    top_dir         => 'top-level directory (default is $PWD)',
    t_dir           => 'the t/ test directory (default is $top_dir/t)',
    t_conf          => 'the conf/ test directory (default is $t_dir/conf)',
    t_logs          => 'the logs/ test directory (default is $t_dir/logs)',
+   t_pid_file      => 'location of the pid file (default is $t_logs/httpd.pid)',
    t_conf_file     => 'test httpd.conf file (default is $t_conf/httpd.conf)',
    src_dir         => 'source directory to look for mod_foos.so',
    serverroot      => 'ServerRoot (default is $t_dir)',
@@ -103,7 +92,7 @@ my %vars_to_env = (
 );
 
 my %filepath_conf_opts = map { $_ => 1 }
-    qw(top_dir t_dir t_conf t_logs t_conf_file src_dir serverroot
+    qw(top_dir t_dir t_conf t_logs t_pid_file t_conf_file src_dir serverroot
        documentroot bindir sbindir httpd apxs httpd_conf httpd_conf_extra
        perlpod sslca libmodperl);
 
@@ -164,7 +153,7 @@ sub passenv {
 sub passenv_makestr {
     my @vars;
 
-    for (keys %passenv) {
+    for (sort keys %passenv) {
         push @vars, "$_=\$($_)";
     }
 
@@ -179,10 +168,6 @@ sub modperl_build_config {
 
     my $server = ref $self ? $self->server : new_test_server();
 
-    # we don't want to get mp2 preconfigured data in order to be able
-    # to get the interactive tests running.
-    return undef if $ENV{APACHE_TEST_INTERACTIVE_CONFIG_TEST};
-
     # we can't do this if we're using httpd 1.3.X
     # even if mod_perl2 is installed on the box
     # similarly, we shouldn't be loading mp2 if we're not
@@ -193,6 +178,7 @@ sub modperl_build_config {
         eval {
             require Apache2::Build;
         } or return;
+
         return Apache2::Build->build_config;
     }
 
@@ -262,10 +248,6 @@ sub new {
         }
     }
 
-    # custom config options from Apache::TestConfigData
-    # again, this should force reconfiguration
-    custom_config_add_conf_opts($args);
-
     my $self = bless {
         clean => {},
         vhosts => {},
@@ -294,8 +276,16 @@ sub new {
     $self->add_inc;
 
     #help to find libmodperl.so
-    my $src_dir = catfile $vars->{top_dir}, qw(src modules perl);
-    $vars->{src_dir}      ||= $src_dir if -d $src_dir;
+    unless ($vars->{src_dir}) {
+        my $src_dir = catfile $vars->{top_dir}, qw(.. src modules perl);
+
+        if (-d $src_dir) {
+	        $vars->{src_dir} = $src_dir;
+    	} else {
+	        $src_dir = catfile $vars->{top_dir}, qw(src modules perl);
+	        $vars->{src_dir} = $src_dir if -d $src_dir;
+    	}
+    }
 
     $vars->{t_dir}        ||= catfile $vars->{top_dir}, 't';
     $vars->{serverroot}   ||= $vars->{t_dir};
@@ -308,6 +298,7 @@ sub new {
     $vars->{sslcaorg}     ||= 'asf';
     $vars->{t_logs}       ||= catfile $vars->{serverroot}, 'logs';
     $vars->{t_conf_file}  ||= catfile $vars->{t_conf},   'httpd.conf';
+    $vars->{t_pid_file}   ||= catfile $vars->{t_logs},   'httpd.pid';
 
     if (WINFU) {
         for (keys %$vars) {
@@ -385,33 +376,22 @@ sub httpd_config {
             # not sure what else could go wrong but we can't continue
             die "something is wrong, mod_perl 2.0 build should have " .
                 "supplied all the needed information to run the tests. " .
-                "Please post lib/Apache/BuildConfig.pm along with the " .
+                "Please post lib/Apache2/BuildConfig.pm along with the " .
                 "bug report";
         }
 
-        if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
-            error "You specified APACHE_TEST_NO_STICKY_PREFERENCES=1 " .
-                "in which case you must explicitly specify -httpd " .
-                "and/or -apxs options";
-            Apache::TestRun::exit_perl(0);
-        }
-
         $self->clean(1);
-        # this method restarts the whole program via exec
-        # so it never returns
-        $self->custom_config_first_time($self->{vars});
+
+        error "You must explicitly specify -httpd and/or -apxs options, " .
+            "or set \$ENV{APACHE_TEST_HTTPD} and \$ENV{APACHE_TEST_APXS}, " .
+            "or set your \$PATH to include the httpd and apxs binaries.";
+        Apache::TestRun::exit_perl(1);
+
     }
     else {
         debug "Using httpd: $vars->{httpd}";
     }
 
-    # if we have gotten that far we know at least about the location
-    # of httpd and or apxs, so let's save it if we haven't saved any
-    # custom configs yet
-    unless (custom_config_exists()) {
-        $self->custom_config_save($self->{vars});
-    }
-
     $self->inherit_config; #see TestConfigParse.pm
     $self->configure_httpd_eapi; #must come after inherit_config
 
@@ -424,7 +404,7 @@ sub httpd_config {
 
     $self->{server}->post_config;
 
-    $self;
+    return $self;
 }
 
 sub default_module {
@@ -578,6 +558,7 @@ sub massage_config_args {
             }
         }
         else {
+            $data=~s/\n(?!\z)/\n    /g;
             $args .= "    $data";
         }
         $args .= "</$directive>\n";
@@ -631,7 +612,8 @@ sub add_config_hooks_run {
 
     for (@{ $self->{$where} }) {
         $self->replace;
-        print $out "$_\n";
+        s/\n?$/\n/;
+        print $out "$_";
     }
 }
 
@@ -684,7 +666,7 @@ sub default_user {
         }
     }
 
-    $user;
+    return $user;
 }
 
 sub default_serveradmin {
@@ -701,7 +683,18 @@ sub default_apxs {
         return $build_config->{MP_APXS};
     }
 
-    $ENV{APACHE_TEST_APXS};
+    if ($ENV{APACHE_TEST_APXS}) {
+        return $ENV{APACHE_TEST_APXS};
+    }
+
+    # look in PATH as a last resort
+    if (my $apxs = which('apxs')) {
+        return $apxs;
+    } elsif ($apxs = which('apxs2')) {
+        return $apxs;
+    }
+    
+    return;
 }
 
 sub default_httpd {
@@ -714,11 +707,35 @@ sub default_httpd {
             for my $bindir (qw(bin sbin)) {
                 my $httpd = catfile $p, $bindir, $vars->{target};
                 return $httpd if -e $httpd;
+                # The executable on Win32 in Apache/2.2 is httpd.exe,
+                # so try that if Apache.exe doesn't exist
+                if (WIN32) {
+                    $httpd = catfile $p, $bindir, 'httpd.EXE';
+                    if (-e $httpd) {
+                        $vars->{target} = 'httpd.EXE';
+                        return $httpd;
+                    }
+                }
             }
         }
     }
 
-    $ENV{APACHE_TEST_HTTPD};
+    if ($ENV{APACHE_TEST_HTTPD}) {
+        return $ENV{APACHE_TEST_HTTPD};
+    }
+
+    # look in PATH as a last resort
+    if (my $httpd = which('httpd')) {
+        return $httpd;
+    } elsif ($httpd = which('httpd2')) {
+        return $httpd;
+    } elsif ($httpd = which('apache')) {
+        return $httpd;
+    } elsif ($httpd = which('apache2')) {
+        return $httpd;
+    }
+    
+    return;
 }
 
 my $localhost;
@@ -750,7 +767,7 @@ my $port_memoized;
 sub select_first_port {
     my $self = shift;
 
-    my $port ||= $port_memoized || $ENV{APACHE_TEST_PORT} 
+    my $port ||= $port_memoized || $ENV{APACHE_TEST_PORT}
         || $self->{vars}{port} || DEFAULT_PORT;
 
     # memoize
@@ -859,12 +876,14 @@ sub find_apache_module {
 my %warn_style = (
     html    => sub { "<!-- @_ -->" },
     c       => sub { "/* @_ */" },
+    php     => sub { "<?php /* \n@_ \n*/ ?>" },
     default => sub { join '', grep {s/^/\# /gm} @_ },
 );
 
 my %file_ext = (
     map({$_ => 'html'} qw(htm html)),
     map({$_ => 'c'   } qw(c h)),
+    map({$_ => 'php' } qw(php)),
 );
 
 # return the passed file's extension or '' if there is no one
@@ -886,9 +905,11 @@ sub warn_style_sub_ref {
 sub genwarning {
     my($self, $filename, $from_filename) = @_;
     return unless $filename;
+    my $time = scalar localtime;
     my $warning = "WARNING: this file is generated";
     $warning .= " (from $from_filename)" if defined $from_filename;
     $warning .= ", do not edit\n";
+    $warning .= "generated on $time\n";
     $warning .= calls_trace();
     return $self->warn_style_sub_ref($filename)->($warning);
 }
@@ -1024,8 +1045,8 @@ sub write_perlscript {
 
     my $fh = $self->genfile($file, undef, 1);
 
-    # shebang
-    print $fh "#!$Config{perlpath}\n";
+    my $shebang = make_shebang();
+    print $fh $shebang;
 
     $self->genfile_warning($file, undef, $fh);
 
@@ -1035,6 +1056,23 @@ sub write_perlscript {
     chmod 0755, $file;
 }
 
+sub make_shebang {
+    # if perlpath is longer than 62 chars, some shells on certain
+    # platforms won't be able to run the shebang line, so when seeing
+    # a long perlpath use the eval workaround.
+    # see: http://en.wikipedia.org/wiki/Shebang
+    # http://homepages.cwi.nl/~aeb/std/shebang/
+    my $shebang = length $Config{perlpath} < 62
+        ? "#!$Config{perlpath}\n"
+        : <<EOI;
+$Config{'startperl'}
+    eval 'exec $Config{perlpath} -S \$0 \${1+"\$@"}'
+        if \$running_under_some_shell;
+EOI
+
+    return $shebang;
+}
+
 sub cpfile {
     my($self, $from, $to) = @_;
     File::Copy::copy($from, $to);
@@ -1086,7 +1124,7 @@ sub clean {
     $self->cmodules_clean;
     $self->sslca_clean;
 
-    for (keys %{ $self->{clean}->{files} }) {
+    for (sort keys %{ $self->{clean}->{files} }) {
         if (-e $_) {
             debug "unlink $_";
             unlink $_;
@@ -1111,6 +1149,10 @@ sub clean {
     }
 }
 
+my %special_tokens = (
+    nextavailableport => sub { shift->server->select_next_port }
+);
+
 sub replace {
     my $self = shift;
     my $file = $Apache::TestConfig::File
@@ -1118,9 +1160,15 @@ sub replace {
 
     s[@(\w+)@]
      [ my $key = lc $1;
-      exists $self->{vars}->{$key}
-      ? $self->{vars}->{$key}
-      : die "invalid token: \@$1\@ $file\n";
+       if (my $callback = $special_tokens{$key}) {
+           $self->$callback;
+       }
+       elsif (exists $self->{vars}->{$key}) {
+           $self->{vars}->{$key};
+       }
+       else {
+           die "invalid token: \@$1\@ $file\n";
+       }
      ]ge;
 }
 
@@ -1193,7 +1241,9 @@ sub parse_vhost {
         @out_config = ([Listen => '0.0.0.0:' . $port]);
 
         if ($self->{vhosts}->{$module}->{namebased}) {
-            push @out_config => [NameVirtualHost => "*:$port"];
+            push @out_config => ["<IfVersion < 2.3.11>\n".
+                                 "${indent}${indent}NameVirtualHost"
+                                 => "*:$port\n${indent}</IfVersion>"];
         }
     }
 
@@ -1358,7 +1408,7 @@ sub extra_conf_files_needing_update {
     finddepth(sub {
         return unless /\.in$/;
         (my $generated = $File::Find::name) =~ s/\.in$//;
-        push @need_update, $generated 
+        push @need_update, $generated
             unless -e $generated && -M $generated < -M $File::Find::name;
     }, $self->{vars}->{t_conf});
 
@@ -1549,7 +1599,7 @@ sub generate_httpd_conf {
 
     $self->preamble_run($out);
 
-    for my $name (qw(user group)) { #win32/cygwin do not support
+    for my $name (qw(user group)) { #win32
         if ($vars->{$name}) {
             print $out qq[\u$name    "$vars->{$name}"\n];
         }
@@ -1570,7 +1620,7 @@ sub generate_httpd_conf {
     # but wasn't included in the system-wide httpd.conf
 
     print $out "<IfModule mod_alias.c>\n";
-    for (keys %aliases) {
+    for (sort keys %aliases) {
         next unless $vars->{$aliases{$_}};
         print $out "    Alias /getfiles-$_ $vars->{$aliases{$_}}\n";
     }
@@ -1603,7 +1653,7 @@ sub need_reconfiguration {
     # if httpd.conf is older than executable
     push @reasons,
         "$exe is newer than $vars->{t_conf_file}"
-            if -e $exe && 
+            if -e $exe &&
                -e $vars->{t_conf_file} &&
                -M $exe < -M $vars->{t_conf_file};
 
@@ -1622,11 +1672,22 @@ sub need_reconfiguration {
     # last run and thus avoid the reconfiguration?
     {
         my $passenv = passenv();
-        if (my @env_vars = grep { $ENV{$_} } keys %$passenv) {
+        if (my @env_vars = sort grep { $ENV{$_} } keys %$passenv) {
             push @reasons, "environment variables (@env_vars) are set";
         }
     }
 
+    # if the generated config was created with a version of Apache-Test
+    # less than the current version
+    {
+      my $current = Apache::Test->VERSION;
+      my $config  = $self->{apache_test_version};
+
+      if (! $config || $config < $current) {
+          push @reasons, "configuration generated with old Apache-Test";
+      }
+    }
+
     return @reasons;
 }
 
@@ -1713,14 +1774,16 @@ sub apxs {
     $self->{_apxs}{$q};
 }
 
-# Temporarily untaint PATH
+# return an untainted PATH
 sub untaint_path {
     my $path = shift;
+    return '' unless defined $path;
     ($path) = ( $path =~ /(.*)/ );
     # win32 uses ';' for a path separator, assume others use ':'
     my $sep = WIN32 ? ';' : ':';
     # -T disallows relative and empty directories in the PATH
-    return join $sep, grep !/^(\.|$)/, split /$sep/, $path;
+    return join $sep, grep File::Spec->file_name_is_absolute($_),
+        grep length($_), split /$sep/, $path;
 }
 
 sub pop_dir {
@@ -1747,12 +1810,15 @@ sub add_inc {
     my $apache_test_dir = catdir $top_dir, "Apache-Test";
     unshift @dirs, $apache_test_dir if -d $apache_test_dir;
 
+    lib::->import(@dirs);
+
     if ($ENV{APACHE_TEST_LIVE_DEV}) {
+        # add lib/ in a separate call to ensure that it'll end up on
+        # top of @INC
         my $lib_dir = catdir $top_dir, "lib";
-        push @dirs, $lib_dir if -d $lib_dir;
+        lib::->import($lib_dir) if -d $lib_dir;
     }
 
-    lib::->import(@dirs);
     #print join "\n", "add_inc", @INC, "";
 }
 
@@ -1804,6 +1870,9 @@ sub save {
 
     return unless $self->{save};
 
+    # add in the Apache-Test version for later comparisions
+    $self->{apache_test_version} = Apache::Test->VERSION;
+
     my $name = 'apache_test_config';
     my $file = catfile $self->{vars}->{t_conf}, "$name.pm";
     my $fh = $self->genfile($file);
@@ -1839,7 +1908,7 @@ sub as_string {
         $cfg .= qx{$command};
 
         $cfg .= ldd_as_string($httpd);
-    } 
+    }
     else {
         $cfg .= "\n\n*** The httpd binary was not found\n";
     }
@@ -1874,7 +1943,7 @@ sub ldd_as_string {
 
     return $cfg;
 }
-  
+
 # make a string suitable for feed to shell calls (wrap in quotes and
 # escape quotes)
 sub shell_ready {
@@ -1883,510 +1952,6 @@ sub shell_ready {
     return qq["$arg"];
 }
 
-### Permanent custom configuration functions ###
-
-# determine which configuration file Apache/TestConfigData.pm to use
-# (as there could be several). The order searched is:
-# 1) $ENV{HOME}/.apache-test/
-# 2) in @INC
-my $custom_config_path;
-sub custom_config_path {
-
-    return $custom_config_path if $custom_config_path;
-
-    my @inc  = ();
-
-    # XXX $ENV{HOME} isn't propagated in mod_perl
-    push @inc, catdir $ENV{HOME}, '.apache-test' if $ENV{HOME};
-
-    push @inc, @INC;
-
-    for (@inc) {
-        my $candidate = File::Spec->rel2abs(catfile $_, CUSTOM_CONFIG_FILE);
-        next unless -e $candidate;
-        # launder for -T
-        ($candidate) = $candidate =~ /^(.*)/;
-        return $custom_config_path = $candidate;
-    }
-
-    return '';
-}
-
-# tries to nuke all occurences of custom config
-# used by things outside the A-T test suite
-sub custom_config_nuke {
-    my $cwd = fastcwd();
-
-    # 1) create a fake empty (blib/)?lib/Apache/TestConfigData.pm
-    # (don't delete it since it may mess up with MakeMaker)
-    my $path = catfile $cwd, "lib", Apache::TestConfig::CUSTOM_CONFIG_FILE;
-    # overwrite the empty stub
-    Apache::TestConfig::custom_config_write($path, '') if -e $path;
-
-    $path = catfile $cwd, "blib", "lib",
-        Apache::TestConfig::CUSTOM_CONFIG_FILE;
-    if (-e $path) {
-        my $mode = (stat _)[2];
-        my $mode_new = $mode | 0200;
-        chmod $mode_new, $path;
-        debug  "emptying $path";
-        Apache::TestConfig::custom_config_write($path, '');
-        chmod $mode, $path;
-    }
-
-    # 2) go through @INC = ~/.apache-test and nuke any occurences of
-    #    CUSTOM_CONFIG_FILE
-    my @inc  = ();
-
-    push @inc, catdir $ENV{HOME}, '.apache-test' if $ENV{HOME};
-
-    push @inc, @INC;
-
-    for (@inc) {
-        my $victim = File::Spec->rel2abs(catfile $_, CUSTOM_CONFIG_FILE);
-        next unless -e $victim;
-        debug "unlinking $victim";
-        unlink $victim;
-    }
-}
-
-sub custom_config_exists {
-    # try to load custom config if it wasn't loaded yet (there are
-    # many entry points to this API)
-    custom_config_load();
-
-    # it's enough to check whether we have a custom
-    # config for 'httpd' or 'apxs'.
-    my $httpd = $Apache::TestConfigData::vars->{httpd} || '';
-    return 1 if $httpd && -e $httpd && -x _;
-
-    my $apxs = $Apache::TestConfigData::vars->{apxs} || '';
-    return 1 if $apxs && -e $apxs && -x _;
-
-    return 0;
-}
-
-# to be used only from Apache-Test/Makefile.PL to write the custom
-# configuration module so it'll be copied to blib during 'make' and
-# updated to use custom config data during 'make test' and then
-# installed system-wide via 'make install'
-#
-# it gets written only if the custom configuration didn't exist
-# already
-sub custom_config_file_stub_write {
-
-    return if custom_config_exists();
-
-    # It doesn't matter whether it gets written under modperl-2.0/lib
-    # or Apache-Test/lib root, since Apache::TestRun uses the same
-    # logic and will update that file with real config data, which
-    # 'make install' will then pick and install system-wide. but
-    # remember that $FindBin::Bin is the location of top-level
-    # 'Makefile.PL'
-    require FindBin; # load it only for this particular use
-    my $path = catfile $FindBin::Bin, "lib",
-        Apache::TestConfig::CUSTOM_CONFIG_FILE;
-
-    # write an empty stub
-    Apache::TestConfig::custom_config_write($path, '');
-}
-
-sub custom_config_save {
-    my $self = shift;
-    my $conf_opts = shift;
-
-    if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
-        debug "APACHE_TEST_NO_STICKY_PREFERENCES=1 => " .
-            "skipping save of custom config data";
-        return;
-    }
-
-    my $vars = $self->{vars};
-    my $config_dump = '';
-
-    # minimum httpd and/or apxs needs to be set
-    return 0 unless $vars->{httpd} or $Apache::TestConfigData::vars->{httpd}
-        or          $vars->{apxs}  or $Apache::TestConfigData::vars->{apxs};
-
-    # it doesn't matter how these vars were set (httpd may or may not
-    # get set using the path to apxs, w/o an explicit -httpd value)
-    for (@data_vars_must) {
-        next unless my $var = $vars->{$_} || $conf_opts->{$_};
-        $config_dump .= qq{    '$_' => '$var',\n};
-    }
-
-    # save these vars only if they were explicitly set via command line
-    # options. For example if someone builds A-T as user 'foo', then
-    # installs it as root and we save it, all users will now try to
-    # configure under that user 'foo' which won't quite work.
-    for (@data_vars_opt) {
-        next unless my $var = $conf_opts->{$_};
-        $config_dump .= qq{    '$_' => '$var',\n};
-    }
-
-    if (IS_APACHE_TEST_BUILD) {
-        my $path = catfile $vars->{top_dir}, 'lib', CUSTOM_CONFIG_FILE;
-        # if it doesn't exist, then we already have a global config file
-        # if it does, then we have need to update it and its blib/ copy
-        if (-e $path and custom_config_path_is_writable($path)) {
-            custom_config_write($path, $config_dump);
-            # also update blib/lib, since usually that's the one that
-            # appears in @INC when t/TEST is run. and it won't be
-            # synced with blib/ unless 'make' was run
-            my $blib_path = catfile $vars->{top_dir},
-                'blib', 'lib', CUSTOM_CONFIG_FILE;
-            if (-e $blib_path and custom_config_path_is_writable($blib_path)) {
-                custom_config_write($blib_path, $config_dump);
-            }
-            return 1;
-        }
-    }
-
-    my $path;
-    if ($path = custom_config_path() ) {
-        # do nothing, the config file already exists (global)
-        debug "Found custom config '$path'";
-    }
-    elsif (File::Spec->file_name_is_absolute(__FILE__)) {
-        # next try a global location, as if it was configured before
-        # Apache::Test's 'make install' (install in the same dir as
-        # Apache/TestRun.pm)
-        # if the filename is not absolute that means that we are still
-        # in Apache-Test build (could just test for IS_APACHE_TEST_BUILD)
-        my $base = dirname dirname __FILE__;
-        $path = catdir $base, CUSTOM_CONFIG_FILE;
-    }
-
-    # check whether we can write to the directory of the chosen path
-    # (e.g. root-owned directory)
-    if ($path and custom_config_path_is_writable($path)) {
-        custom_config_write($path, $config_dump);
-        return 1;
-    }
-    # if we have no writable path yet, try to use ~
-    elsif ($ENV{HOME}) {
-        $path = catfile $ENV{HOME}, '.apache-test', CUSTOM_CONFIG_FILE;
-        if ($path and custom_config_path_is_writable($path)) {
-            custom_config_write($path, $config_dump);
-            return 1;
-        }
-    }
-
-    # XXX: should we croak since we failed to write config
-    error "Failed to find a config file to save the custom " .
-        "configuration in";
-    return 0;
-}
-
-sub custom_config_path_is_writable {
-    my $path = shift;
-
-    return 0 unless $path;
-
-    my $file_created    = '';
-    my $top_dir_created = '';
-    # first make sure that the file is writable if it exists
-    # already (it might be non-writable if installed via EU::MM or in
-    # blib/)
-    if (-e $path) {
-        my $mode = (stat _)[2];
-        $mode |= 0200;
-        chmod $mode, $path; # it's ok if we fail
-        # keep it writable if we have changed it from not being one
-        # so that custom_config_save will be able to just overwrite it
-    }
-    else {
-        my $dir = dirname $path;
-        if ($dir and !-e $dir) {
-            my @dirs = File::Path::mkpath($dir, 0, 0755);
-            # the top level dir to nuke on cleanup if it was created
-            $top_dir_created = shift @dirs if @dirs;
-        }
-        # not really create yet, but will be in the moment
-        $file_created = 1;
-    }
-
-    # try to open for append (even though it may not exist
-    my $fh = Symbol::gensym;
-    if (open $fh, ">>$path") {
-        close $fh;
-        # cleanup if we just created the file
-        unlink $path if $file_created;
-        File::Path::rmtree([$top_dir_created], 0, 0) if $top_dir_created;
-        return 1;
-    }
-
-    return 0;
-}
-
-sub custom_config_write {
-    my($path, $config_dump) = @_;
-
-    my $pkg = << "EOC";
-package Apache::TestConfigData;
-
-use strict;
-use warnings;
-
-\$Apache::TestConfigData::vars = {
-$config_dump
-};
-
-1;
-
-=head1 NAME
-
-Apache::TestConfigData - Configuration file for Apache::Test
-
-=cut
-EOC
-
-    debug "Writing custom config $path";
-    my $dir = dirname $path;
-    File::Path::mkpath($dir, 0, 0755) unless -e $dir;
-    my $fh = Symbol::gensym;
-    open $fh, ">$path" or die "Cannot open $path: $!";
-    print $fh $pkg;
-    close $fh;
-}
-
-sub custom_config_add_conf_opts {
-    my $args = shift;
-
-    return unless $Apache::TestConfigData::vars and 
-        keys %$Apache::TestConfigData::vars;
-
-    debug "overlaying custom config data";
-
-    # the logic is quite complicated with 'httpd' and 'apxs', since
-    # one is enough to run the test suite, and we need to avoid the
-    # situation where both are saved in custom config but only one
-    # (let's say httpd) is overriden by the command line /env var and
-    # a hell may break loose if we take that overriden httpd value and
-    # also use apxs from custom config which could point to a different
-    # server. So if there is an override of apxs or httpd, do not use
-    # the custom config for apxs or httpd.
-    my $vars_must_overriden = grep {
-        $ENV{ $vars_to_env{$_} } || $args->{$_}
-    } @data_vars_must;
-
-    # mod_perl 2.0 build always knows the right httpd location (and
-    # optionally apxs)
-    $vars_must_overriden++ if IS_MOD_PERL_2_BUILD();
-
-    unless ($vars_must_overriden) {
-        for (@data_vars_must) {
-            next unless $Apache::TestConfigData::vars->{$_};
-            $args->{$_} = $Apache::TestConfigData::vars->{$_};
-        }
-    }
-
-    for (@data_vars_opt) {
-        next unless $Apache::TestConfigData::vars->{$_};
-        # env vars override custom config
-        my $env_value = $ENV{ $vars_to_env{$_} };
-        next unless defined $env_value and length $env_value;
-        $args->{$_} ||= $Apache::TestConfigData::vars->{$_};
-    }
-}
-
-my $custom_config_loaded = 0;
-sub custom_config_load {
-
-    if ($ENV{APACHE_TEST_NO_STICKY_PREFERENCES}) {
-        debug "APACHE_TEST_NO_STICKY_PREFERENCES=1 => " .
-            "skipping load of custom config data";
-        return;
-    }
-
-    if ($ENV{APACHE_TEST_INTERACTIVE_CONFIG_TEST}) {
-        debug "APACHE_TEST_INTERACTIVE_CONFIG_TEST=1 => " .
-            "skipping load of custom config data";
-        return;
-    }
-
-    return if $custom_config_loaded;
-
-    if (my $custom_config_path = custom_config_path()) {
-        debug "loading custom config data from: '$custom_config_path'";
-        $custom_config_loaded++;
-        require $custom_config_path;
-    }
-    else {
-        debug "no custom config data was loaded";
-    }
-}
-
-sub custom_config_first_time {
-    my $self = shift;
-    my $conf_opts = shift;
-
-    # we can't prompt when STDIN is not attached to tty, unless we
-    # were told that's it OK via env var (in which case some program
-    # will feed the interactive prompts
-    unless (-t STDIN || $ENV{APACHE_TEST_INTERACTIVE_PROMPT_OK}) {
-        error "STDIN is not attached to tty, skip interactive config";
-        Apache::TestRun::skip_test_suite();
-    }
-
-    my $vars = $self->{vars};
-
-    print qq[
-
-We are now going to configure the Apache-Test framework.
-This configuration process needs to be done only once.
-
-];
-
-    print qq[
-
-First we need to know where the 'httpd' executable is located.
-If you have more than one Apache server is installed, make sure
-you supply the path to the one you are going to use for testing.
-You can always override this setting at run time via the '-httpd'
-option. For example:
-
-  % t/TEST -httpd /path/to/alternative/httpd
-
-or via the environment variable APACHE_TEST_HTTPD. For example:
-
-  % APACHE_TEST_HTTPD=/path/to/alternative/httpd t/TEST
-
-If for some reason you want to skip the test suite, type: skip
-];
-
-    {
-        my %choices = ();
-        my @tries = qw(httpd httpd2);
-        # Win32 uses Apache not apache
-        push @tries, WIN32 ? qw(Apache) : qw(apache);
-        for (grep defined $_,
-             map({ catfile $vars->{$_}, $vars->{target} } qw(sbindir bindir)),
-             $self->default_httpd, which($vars->{target}),
-             $ENV{APACHE}, $ENV{APACHE2},
-             $ENV{APACHE_TEST_HTTPD}, $ENV{APACHE_TEST_HTTPD2},
-             map {which($_)} @tries) {
-            $choices{$_}++ if -e $_ && -x _;
-        }
-        my $optional = 0;
-        my $wanted = 'httpd';
-        $vars->{$wanted} = 
-            _custom_config_prompt_path($wanted, \%choices, $optional);
-    }
-
-    print qq[
-
-Next we need to know where the 'apxs' script is located. This script
-provides a lot of information about the apache installation, and makes
-it easier to find things. However it's not available on all platforms,
-therefore it's optional.
-
-If you don't have it installed it's not a problem. Just press Enter.
-
-Notice that if you have Apache 2.x installed that script could be
-called as 'apxs2'.
-
-If you have more than one Apache server is installed, make sure you
-supply the path to the apxs script you are going to use for testing.
-You can always override this setting at run time via the '-apxs'
-option. For example:
-
-  % t/TEST -apxs /path/to/alternative/apxs
-
-or via the environment variable APACHE_TEST_APXS. For example:
-
-  % APACHE_TEST_APXS=/path/to/alternative/apxs t/TEST
-
-];
-    {
-        my %choices = ();
-        for (grep defined $_,
-             map({ catfile $vars->{$_}, 'apxs' } qw(sbindir bindir)),
-             $self->default_apxs,
-             $ENV{APXS},  $ENV{APACHE_TEST_APXS},  which('apxs'),
-             $ENV{APXS2}, $ENV{APACHE_TEST_APXS2}, which('apxs2')) {
-            $choices{$_}++ if -e $_ && -x _;
-        }
-        my $optional = 1;
-        my $wanted = 'apxs';
-        $vars->{$wanted} = 
-            _custom_config_prompt_path($wanted, \%choices, $optional);
-    }
-
-    $self->custom_config_save($conf_opts);
-
-    # we probably could reconfigure on the fly ($self->configure), but
-    # the problem is various cached data which won't be refreshed. so
-    # the simplest is just to restart the run from scratch
-    Apache::TestRun::rerun($vars);
-}
-
-sub _custom_config_prompt_path {
-    my($wanted, $rh_choices, $optional) = @_;
-
-    my $ans;
-    my $default = '';
-    my $optional_str = $optional ? " (optional)" : '';
-    my $prompt =
-        "\nPlease provide a full path to$optional_str '$wanted' executable";
-
-    my @choices = ();
-    if (%$rh_choices) {
-        $prompt .= " or choose from the following options:\n\n";
-        my $c = 0;
-        for (sort keys %$rh_choices) {
-            $c++;
-            $prompt .= "    [$c] $_\n";
-            push @choices, $_;
-        }
-        $prompt .= " \n";
-        $default = 1; # a wild guess
-    }
-    else {
-        $prompt .= ":\n\n";
-    }
-
-    while (1) {
-        $ans = ExtUtils::MakeMaker::prompt($prompt, $default);
-
-        # strip leading/closing spaces
-        $ans =~ s/^\s*|\s*$//g;
-
-        # convert the item number to the path
-        if ($ans =~ /^(\d+)$/) {
-            if ($1 > 0 and $choices[$1-1]) {
-                $ans = $choices[$1-1];
-            }
-            else {
-                warn "The choice '$ans' doesn't exist\n";
-                next;
-            }
-        }
-
-        if ($optional) {
-            return '' unless $ans;
-        }
-
-        # stop the test suite without an error (so automatic tools
-        # like CPAN.pm will be able to continue)
-        if (lc($ans) eq 'skip' && !$optional) {
-            Apache::TestRun::skip_test_suite();
-            next; # in case they change their mind
-        }
-
-        unless (File::Spec->file_name_is_absolute($ans)) {
-            warn "The path '$ans' is not an absolute path. " .
-                "Please specify an absolute path\n";
-            next;
-        }
-
-        warn("'$ans' doesn't exist.\n"),     next unless -e $ans;
-        warn("'$ans' is not a file.\n"),     next unless -f _;
-        warn("'$ans' is not executable.\n"), next unless -x _;
-
-        return $ans;
-    }
-}
 
 1;
 
@@ -2526,15 +2091,43 @@ configuration phase (C<t/TEST -config>, C<Apache::Test> will
 automatically unshift the I<project/lib> directory into C<@INC>, via
 the autogenerated I<t/conf/modperl_inc.pl> file.
 
-=head2 APACHE_TEST_INTERACTIVE_PROMPT_OK
 
-Normally interactive prompts aren't run when STDIN is not attached to
-a tty. But sometimes there is a program that can answer the prompts
-(e.g. when testing A-T itself). If this variable is true the
-interactive config won't be skipped (if needed).
+=head1 Special Placeholders
 
+When generating configuration files from the I<*.in> templates,
+special placeholder variables get substituted. To embed a placeholder
+use the C<@foo@> syntax. For example in I<extra.conf.in> you can
+write:
 
+  Include @ServerRoot@/conf/myconfig.conf
 
+When I<extra.conf> is generated, C<@ServerRoot@> will get replaced
+with the location of the server root.
+
+Placeholders are case-insensitive.
+
+Available placeholders:
+
+=head2 Configuration Options
+
+All configuration variables that can be passed to C<t/TEST>, such as
+C<MaxClients>, C<DocumentRoot>, C<ServerRoot>, etc. To see the
+complete list run:
+
+  % t/TEST --help
+
+and you will find them in the C<configuration options> sections.
+
+=head2 NextAvailablePort
+
+Every time this placeholder is encountered it'll be replaced with the
+next available port. This is very useful if you need to allocate a
+special port, but not hardcode it. Later when running:
+
+  % t/TEST -port=select
+
+it's possible to run several concurrent test suites on the same
+machine, w/o having port collisions.
 
 =head1 AUTHOR
 
@@ -2551,10 +2144,16 @@ Listen     0.0.0.0:@Port@
 ServerRoot   "@ServerRoot@"
 DocumentRoot "@DocumentRoot@"
 
-PidFile     @t_logs@/httpd.pid
+PidFile     @t_pid_file@
 ErrorLog    @t_logs@/error_log
 LogLevel    debug
 
+<IfModule mod_version.c>
+<IfVersion > 2.4.1>
+    DefaultRunTimeDir "@t_logs@"
+</IfVersion>
+</IfModule>
+
 <IfModule mod_log_config.c>
     TransferLog @t_logs@/access_log
 </IfModule>
@@ -2576,7 +2175,11 @@ HostnameLookups Off
 </Directory>
 
 <IfModule @THREAD_MODULE@>
+<IfModule mod_version.c>
+<IfVersion < 2.3.4>
     LockFile             @t_logs@/accept.lock
+</IfVersion>
+</IfModule>
     StartServers         1
     MinSpareThreads      @MinClients@
     MaxSpareThreads      @MinClients@
@@ -2586,7 +2189,11 @@ HostnameLookups Off
 </IfModule>
 
 <IfModule perchild.c>
+<IfModule mod_version.c>
+<IfVersion < 2.3.4>
     LockFile             @t_logs@/accept.lock
+</IfVersion>
+</IfModule>
     NumServers           1
     StartThreads         @MinClients@
     MinSpareThreads      @MinClients@
@@ -2596,7 +2203,11 @@ HostnameLookups Off
 </IfModule>
 
 <IfModule prefork.c>
+<IfModule mod_version.c>
+<IfVersion < 2.3.4>
     LockFile             @t_logs@/accept.lock
+</IfVersion>
+</IfModule>
     StartServers         @MinClients@
     MinSpareServers      @MinClients@
     MaxSpareServers      @MinClients@
