@@ -1,8 +1,10 @@
-# Copyright 2001-2005 The Apache Software Foundation
-#
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# please insert nothing before this line: -*- mode: cperl; cperl-indent-level: 4; cperl-continued-statement-offset: 4; indent-tabs-mode: nil -*-
+# Licensed to the Apache Software Foundation (ASF) under one or more
+# contributor license agreements.  See the NOTICE file distributed with
+# this work for additional information regarding copyright ownership.
+# The ASF licenses this file to You under the Apache License, Version 2.0
+# (the "License"); you may not use this file except in compliance with
+# the License.  You may obtain a copy of the License at
 #
 #     http://www.apache.org/licenses/LICENSE-2.0
 #
@@ -148,7 +150,12 @@ EOI
     require Apache2::Connection;
     require APR::SockAddr;
     require Socket;
-    my $orig_sub = *Apache2::Connection::remote_addr{CODE};
+    my $orig_sub;
+    if (defined *Apache2::Connection::client_addr{CODE}) { # httpd-2.4
+        $orig_sub = *Apache2::Connection::client_addr{CODE};
+    } else { # httpd-2.2
+        $orig_sub = *Apache2::Connection::remote_addr{CODE};
+    }
     *Apache2::Connection::remote_addr = sub {
         my $c = shift;
         if (@_) {
@@ -309,7 +316,7 @@ sub request {
 
 package Apache::Server;
 # XXX: is that good enough? see modperl/src/modules/perl/mod_perl.c:367
-our $CWD = Apache2::ServerUtil::server_root;
+our $CWD = Apache2::ServerUtil::server_root();
 
 our $AddPerlVersion = 1;
 
@@ -320,6 +327,10 @@ sub warn {
 
 package Apache;
 
+sub request {
+    return Apache2::compat::request(@_);
+}
+
 sub unescape_url_info {
     my ($class, $string) = @_;
     Apache2::URI::unescape_url($string);
@@ -343,7 +354,7 @@ sub server_root_relative {
          return File::Spec->catfile(@_);
     }
     else {
-        File::Spec->catfile(Apache2::ServerUtil::server_root, @_);
+        File::Spec->catfile(Apache2::ServerUtil::server_root(), @_);
     }
 }
 
@@ -485,18 +496,18 @@ sub table_get_set {
     my ($key, $value) = @_;
 
     if (1 == @_) {
-        return wantarray() 
+        return wantarray()
             ?       ($table->get($key))
             : scalar($table->get($key));
     }
     elsif (2 == @_) {
         if (defined $value) {
-            return wantarray() 
+            return wantarray()
                 ?        ($table->set($key, $value))
                 :  scalar($table->set($key, $value));
         }
         else {
-            return wantarray() 
+            return wantarray()
                 ?       ($table->unset($key))
                 : scalar($table->unset($key));
         }
@@ -512,21 +523,21 @@ sub table_get_set {
 
 sub header_out {
     my $r = shift;
-    return wantarray() 
+    return wantarray()
         ?       ($r->table_get_set(scalar($r->headers_out), @_))
         : scalar($r->table_get_set(scalar($r->headers_out), @_));
 }
 
 sub header_in {
     my $r = shift;
-    return wantarray() 
+    return wantarray()
         ?       ($r->table_get_set(scalar($r->headers_in), @_))
         : scalar($r->table_get_set(scalar($r->headers_in), @_));
 }
 
 sub err_header_out {
     my $r = shift;
-    return wantarray() 
+    return wantarray()
         ?       ($r->table_get_set(scalar($r->err_headers_out), @_))
         : scalar($r->table_get_set(scalar($r->err_headers_out), @_));
 }
@@ -595,7 +606,7 @@ sub content {
 
 sub server_root_relative {
     my $r = shift;
-    File::Spec->catfile(Apache2::ServerUtil::server_root, @_);
+    File::Spec->catfile(Apache2::ServerUtil::server_root(), @_);
 }
 
 sub clear_rgy_endav {
