@@ -1,8 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -60,7 +61,7 @@ static apr_status_t modperl_shutdown(void *data)
 
     handles = modperl_xs_dl_handles_get(aTHX);
 
-    MP_TRACE_i(MP_FUNC, "destroying interpreter=0x%lx\n",
+    MP_TRACE_i(MP_FUNC, "destroying interpreter=0x%lx",
                (unsigned long)perl);
 
     modperl_perl_destruct(perl);
@@ -164,8 +165,7 @@ static void set_taint_var(PerlInterpreter *perl)
     dTHXa(perl);
 
 /* 5.7.3+ has a built-in special ${^TAINT}, backport it to 5.6.0+ */
-#if PERL_REVISION == 5 && \
-    (PERL_VERSION == 6 || (PERL_VERSION == 7 && PERL_SUBVERSION < 3))
+#if MP_PERL_VERSION_AT_MOST(5, 7, 2)
     {
         GV *gv = gv_fetchpv("\024AINT", GV_ADDMULTI, SVt_IV);
         sv_setiv(GvSV(gv), PL_tainting);
@@ -215,13 +215,15 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
     {
         server_rec *base_server = modperl_global_get_server_rec();
         const char *desc = modperl_server_desc(s, p);
-        if (base_server == s) {        
+        if (base_server == s) {
+            MP_init_status = 1; /* temporarily reset MP_init_status */
             MP_TRACE_i(MP_FUNC,
-                       "starting the parent perl for the base server\n", desc);
+                       "starting the parent perl for the base server", desc);
+            MP_init_status = 2;
         }
         else {
             MP_TRACE_i(MP_FUNC,
-                       "starting the parent perl for vhost %s\n", desc);
+                       "starting the parent perl for vhost %s", desc);
         }
     }
 #endif
@@ -262,10 +264,10 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 
     /* suspend END blocks to be run at server shutdown */
     endav = PL_endav;
-    PL_endav = Nullav;
+    PL_endav = (AV *)NULL;
 
 /* This was fixed in 5.9.0/5.8.1 (17775), but won't compile after 19122 */
-#if PERL_REVISION == 5 && PERL_VERSION == 8 && PERL_SUBVERSION == 0 && \
+#if MP_PERL_VERSION(5, 8, 0) && \
     defined(USE_REENTRANT_API) && defined(HAS_CRYPT_R) && defined(__GLIBC__)
     /* workaround perl5.8.0/glibc bug */
     PL_reentrant_buffer->_crypt_struct.current_saltbits = 0;
@@ -292,7 +294,7 @@ PerlInterpreter *modperl_startup(server_rec *s, apr_pool_t *p)
 
     set_taint_var(perl);
 
-    MP_TRACE_i(MP_FUNC, "constructed interpreter=0x%lx\n",
+    MP_TRACE_i(MP_FUNC, "constructed interpreter=0x%lx",
                (unsigned long)perl);
 
 #ifdef MP_USE_GTOP
@@ -351,7 +353,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
     const char *vhost = modperl_server_desc(s, p);
 
     if (!scfg) {
-        MP_TRACE_i(MP_FUNC, "server %s has no mod_perl config\n", vhost);
+        MP_TRACE_i(MP_FUNC, "server %s has no mod_perl config", vhost);
         return OK;
     }
 
@@ -359,11 +361,11 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
         base_server = modperl_global_get_server_rec();
     }
 
-    MP_TRACE_i(MP_FUNC, "Init vhost %s: s=0x%lx, base_s=0x%lx\n",
+    MP_TRACE_i(MP_FUNC, "Init vhost %s: s=0x%lx, base_s=0x%lx",
                vhost, s, base_server);
 
     if (base_server == s) {
-        MP_TRACE_i(MP_FUNC, "base server is not vhost, skipping %s\n",
+        MP_TRACE_i(MP_FUNC, "base server is not vhost, skipping %s",
                    vhost);
         return OK;
     }
@@ -379,19 +381,20 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
 #ifdef USE_ITHREADS
 
     if (scfg->mip) {
-        MP_TRACE_i(MP_FUNC, "server %s already initialized\n", vhost);
+        MP_TRACE_i(MP_FUNC, "server %s already initialized", vhost);
         return OK;
     }
 
     /* the base server could have mod_perl callbacks disabled, but it
      * still needs perl to drive the vhosts */
     if (!MpSrvENABLE(scfg) && s->is_virtual) {
-        MP_TRACE_i(MP_FUNC, "mod_perl disabled for server %s\n", vhost);
+        MP_TRACE_i(MP_FUNC, "mod_perl disabled for server %s", vhost);
         scfg->mip = NULL;
         return OK;
     }
 
     PERL_SET_CONTEXT(perl);
+    modperl_thx_interp_set(perl, base_scfg->mip->parent);
 
 #endif /* USE_ITHREADS */
 
@@ -401,7 +404,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
     if (MpSrvPARENT(scfg)) {
         perl = modperl_startup(s, p);
         MP_TRACE_i(MP_FUNC,
-                   "created parent interpreter for VirtualHost %s\n",
+                   "created parent interpreter for VirtualHost %s",
                    modperl_server_desc(s, p));
     }
     else {
@@ -409,7 +412,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
         /* virtual host w/ +Clone gets its own mip */
         if (MpSrvCLONE(scfg)) {
             modperl_interp_init(s, p, perl);
-        }   
+        }
 #endif
 
         if (!modperl_config_apply_PerlRequire(s, scfg, perl, p)) {
@@ -427,7 +430,7 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
          * need to point to the base mip here if this vhost
          * doesn't have its own
          */
-        MP_TRACE_i(MP_FUNC, "%s mip inherited from %s\n",
+        MP_TRACE_i(MP_FUNC, "%s mip inherited from %s",
                    vhost, modperl_server_desc(base_server, p));
         scfg->mip = base_scfg->mip;
     }
@@ -439,10 +442,10 @@ int modperl_init_vhost(server_rec *s, apr_pool_t *p,
 void modperl_init(server_rec *base_server, apr_pool_t *p)
 {
     server_rec *s;
-    modperl_config_srv_t *base_scfg;
     PerlInterpreter *base_perl;
-
-    base_scfg = modperl_config_srv_get(base_server);
+#if defined(MP_TRACE) || defined(USE_ITHREADS)
+    modperl_config_srv_t *base_scfg = modperl_config_srv_get(base_server);
+#endif
 
     MP_TRACE_d_do(MpSrv_dump_flags(base_scfg,
                                    base_server->server_hostname));
@@ -467,6 +470,7 @@ void modperl_init(server_rec *base_server, apr_pool_t *p)
     /* after other parent perls were started in vhosts, make sure that
      * the context is set to the base_perl */
     PERL_SET_CONTEXT(base_perl);
+    modperl_thx_interp_set(base_perl, base_scfg->mip->parent);
 #endif
 
 }
@@ -476,7 +480,7 @@ static int modperl_post_config_require(server_rec *s, apr_pool_t *p)
     for (; s; s=s->next) {
         MP_dSCFG(s);
         if (!modperl_config_apply_PerlPostConfigRequire(s, scfg, p)) {
-            return FALSE;   
+            return FALSE;
         }
     }
     return TRUE;
@@ -485,13 +489,13 @@ static int modperl_post_config_require(server_rec *s, apr_pool_t *p)
 #ifdef USE_ITHREADS
 static void modperl_init_clones(server_rec *s, apr_pool_t *p)
 {
-    modperl_config_srv_t *base_scfg = modperl_config_srv_get(s);
 #ifdef MP_TRACE
+    modperl_config_srv_t *base_scfg = modperl_config_srv_get(s);
     char *base_name = modperl_server_desc(s, p);
 #endif /* MP_TRACE */
 
     if (!modperl_threaded_mpm()) {
-        MP_TRACE_i(MP_FUNC, "no clones created for non-threaded mpm\n");
+        MP_TRACE_i(MP_FUNC, "no clones created for non-threaded mpm");
         return;
     }
 
@@ -499,9 +503,6 @@ static void modperl_init_clones(server_rec *s, apr_pool_t *p)
         MP_dSCFG(s);
 #ifdef MP_TRACE
         char *name = modperl_server_desc(s, p);
-
-        MP_TRACE_i(MP_FUNC, "PerlInterpScope set to %s for %s\n",
-                   modperl_interp_scope_desc(scfg->interp_scope), name);
 #else
         char *name = NULL;
 #endif /* MP_TRACE */
@@ -510,18 +511,18 @@ static void modperl_init_clones(server_rec *s, apr_pool_t *p)
 #ifdef MP_TRACE
             if (scfg->mip == base_scfg->mip) {
                 MP_TRACE_i(MP_FUNC,
-                           "%s interp pool inherited from %s\n",
+                           "%s interp pool inherited from %s",
                            name, base_name);
             }
             else {
                 MP_TRACE_i(MP_FUNC,
-                           "%s interp pool already initialized\n",
+                           "%s interp pool already initialized",
                            name);
             }
 #endif /* MP_TRACE */
         }
         else {
-            MP_TRACE_i(MP_FUNC, "initializing interp pool for %s\n",
+            MP_TRACE_i(MP_FUNC, "initializing interp pool for %s",
                        name);
             modperl_tipool_init(scfg->mip->tipool);
         }
@@ -533,7 +534,7 @@ void modperl_init_globals(server_rec *s, apr_pool_t *pconf)
 {
     ap_mpm_query(AP_MPMQ_IS_THREADED, &MP_threaded_mpm);
 
-    MP_TRACE_g(MP_FUNC, "mod_perl globals are configured\n");
+    MP_TRACE_g(MP_FUNC, "mod_perl globals are configured");
 
     modperl_global_init_pconf(pconf, pconf);
     modperl_global_init_server_rec(pconf, s);
@@ -553,7 +554,7 @@ static apr_status_t modperl_sys_init(void)
     int argc = 0;
     char **argv = NULL, **env = NULL;
 
-    MP_TRACE_i(MP_FUNC, "mod_perl sys init\n");
+    MP_TRACE_i(MP_FUNC, "mod_perl sys init");
 
     /* not every OS uses those vars in PERL_SYS_INIT3 macro */
     argc = argc; argv = argv; env = env;
@@ -583,13 +584,18 @@ static apr_status_t modperl_sys_init(void)
 
 static apr_status_t modperl_sys_term(void *data)
 {
+    /* PERL_SYS_TERM() needs 'my_perl' as of 5.9.5 */
+#if MP_PERL_VERSION_AT_LEAST(5, 9, 5) && defined(USE_ITHREADS)
+    modperl_cleanup_data_t *cdata = (modperl_cleanup_data_t *)data;
+    PERL_UNUSED_DECL PerlInterpreter *my_perl = cdata == NULL ? NULL : (PerlInterpreter *)cdata->data;
+#endif
     MP_init_status = 0;
     MP_threads_started = 0;
     MP_post_post_config_phase = 0;
 
     MP_PERL_FREE_THREAD_KEY_WORKAROUND;
 
-    MP_TRACE_i(MP_FUNC, "mod_perl sys term\n");
+    MP_TRACE_i(MP_FUNC, "mod_perl sys term");
 
     modperl_env_unload();
 
@@ -600,15 +606,13 @@ static apr_status_t modperl_sys_term(void *data)
     return APR_SUCCESS;
 }
 
-int modperl_hook_init(apr_pool_t *pconf, apr_pool_t *plog, 
+int modperl_hook_init(apr_pool_t *pconf, apr_pool_t *plog,
                       apr_pool_t *ptemp, server_rec *s)
 {
     if (MP_IS_STARTING || MP_IS_RUNNING) {
         return OK;
     }
 
-    MP_TRACE_i(MP_FUNC, "mod_perl hook init\n");
-
     MP_init_status = 1; /* now starting */
 
     modperl_restart_count_inc(s);
@@ -648,6 +652,10 @@ int modperl_is_running(void)
 int modperl_hook_pre_config(apr_pool_t *p, apr_pool_t *plog,
                             apr_pool_t *ptemp)
 {
+#if AP_MODULE_MAGIC_AT_LEAST(20110329,0)
+    ap_reserve_module_slots_directive("PerlLoadModule");
+#endif
+
     /* we can't have PerlPreConfigHandler without first configuring mod_perl */
 
     /* perl 5.8.1+ */
@@ -702,12 +710,12 @@ static int modperl_hook_post_config_last(apr_pool_t *pconf, apr_pool_t *plog,
     }
 #endif
 
-#if PERL_REVISION == 5 && PERL_VERSION < 9
-#define MP_PERL_VERSION_STAMP "Perl/v%vd"
-#else
+#if MP_PERL_VERSION_AT_LEAST(5, 9, 0)
 #define MP_PERL_VERSION_STAMP "Perl/%" SVf
+#else
+#define MP_PERL_VERSION_STAMP "Perl/v%vd"
 #endif
-    
+
     ap_add_version_component(pconf, MP_VERSION_STRING);
     ap_add_version_component(pconf,
                              Perl_form(aTHX_ MP_PERL_VERSION_STAMP,
@@ -732,7 +740,22 @@ static int modperl_hook_create_request(request_rec *r)
 {
     MP_dRCFG;
 
+#ifdef USE_ITHREADS
+    /* XXX: this is necessary to make modperl_interp_pool_select() work
+     * which is used at runtime only to merge dir-configs by
+     * modperl_module_config_merge().
+     *
+     * Since most requests won't need it it would be good to add some logic
+     * (cheaper logic in terms of CPU cycles) to identify those cases and
+     * avoid the hash operation.
+     */
+    MP_TRACE_i(MP_FUNC, "setting userdata MODPERL_R in pool %#lx to %lx",
+               (unsigned long)r->pool, (unsigned long)r);
+    (void)apr_pool_userdata_set((void *)r, "MODPERL_R", NULL, r->pool);
+#endif
+
     modperl_config_req_init(r, rcfg);
+    modperl_config_req_cleanup_register(r, rcfg);
 
     /* set the default for cgi header parsing On as early as possible
      * so $r->content_type in any phase after header_parser could turn
@@ -746,6 +769,12 @@ static int modperl_hook_create_request(request_rec *r)
 
 static int modperl_hook_post_read_request(request_rec *r)
 {
+#ifdef USE_ITHREADS
+    MP_TRACE_i(MP_FUNC, "%s %s:%d%s",
+               r->method, r->connection->local_addr->hostname,
+               r->connection->local_addr->port, r->unparsed_uri);
+#endif
+
     /* if 'PerlOptions +GlobalRequest' is outside a container */
     modperl_global_request_cfg_set(r);
 
@@ -753,7 +782,7 @@ static int modperl_hook_post_read_request(request_rec *r)
 }
 
 static int modperl_hook_header_parser(request_rec *r)
-{    
+{
     /* if 'PerlOptions +GlobalRequest' is inside a container */
     modperl_global_request_cfg_set(r);
 
@@ -830,6 +859,7 @@ void modperl_register_hooks(apr_pool_t *p)
 
 #ifdef USE_ITHREADS
     APR_REGISTER_OPTIONAL_FN(modperl_interp_unselect);
+    APR_REGISTER_OPTIONAL_FN(modperl_thx_interp_get);
 #endif
 
     /* for <IfDefine MODPERL2> and Apache2->define("MODPERL2") */
@@ -890,17 +920,22 @@ void modperl_register_hooks(apr_pool_t *p)
                           NULL, NULL, MODPERL_HOOK_REALLY_REALLY_FIRST);
 
     ap_hook_child_init(modperl_hook_child_init,
-                       NULL, NULL, APR_HOOK_FIRST);
+                       NULL, NULL, MODPERL_HOOK_REALLY_REALLY_FIRST);
 
     modperl_register_handler_hooks();
 }
 
-static const command_rec modperl_cmds[] = {  
+static const command_rec modperl_cmds[] = {
     MP_CMD_SRV_ITERATE("PerlSwitches", switches, "Perl Switches"),
     MP_CMD_DIR_ITERATE("PerlModule", modules, "PerlModule"),
     MP_CMD_DIR_ITERATE("PerlRequire", requires, "PerlRequire"),
     MP_CMD_SRV_ITERATE("PerlConfigRequire", config_requires, "PerlConfigRequire"),
     MP_CMD_SRV_ITERATE("PerlPostConfigRequire", post_config_requires, "PerlPostConfigRequire"),
+#if AP_SERVER_MAJORVERSION_NUMBER>2 || \
+    (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER>=3)
+    MP_CMD_SRV_TAKE2("PerlAddAuthzProvider", authz_provider, "PerlAddAuthzProvider"),
+    MP_CMD_SRV_TAKE2("PerlAddAuthnProvider", authn_provider, "PerlAddAuthnProvider"),
+#endif
     MP_CMD_DIR_ITERATE("PerlOptions", options, "Perl Options"),
     MP_CMD_DIR_ITERATE("PerlInitHandler", init_handlers, "Subroutine name"),
     MP_CMD_DIR_TAKE2("PerlSetVar", set_var, "PerlSetVar"),
@@ -935,8 +970,6 @@ static const command_rec modperl_cmds[] = {
                      "Min number of spare Perl interpreters"),
     MP_CMD_SRV_TAKE1("PerlInterpMaxRequests", interp_max_requests,
                      "Max number of requests per Perl interpreters"),
-    MP_CMD_DIR_TAKE1("PerlInterpScope", interp_scope,
-                     "Scope of a Perl interpreter"),
 #endif
 #ifdef MP_COMPAT_1X
     MP_CMD_DIR_FLAG("PerlSendHeader", send_header,
@@ -951,8 +984,8 @@ static const command_rec modperl_cmds[] = {
                     "Turn on -w switch"),
 #endif
     MP_CMD_ENTRIES,
-    { NULL }, 
-}; 
+    { NULL },
+};
 
 void modperl_response_init(request_rec *r)
 {
@@ -985,7 +1018,7 @@ apr_status_t modperl_response_finish(request_rec *r)
     return modperl_wbucket_flush(rcfg->wbucket, FALSE);
 }
 
-static int modperl_response_handler_run(request_rec *r, int finish)
+static int modperl_response_handler_run(request_rec *r)
 {
     int retval;
 
@@ -997,48 +1030,33 @@ static int modperl_response_handler_run(request_rec *r, int finish)
         r->handler = r->content_type; /* let http_core or whatever try */
     }
 
-    if (finish) {
-        apr_status_t rc = modperl_response_finish(r);
-        if (rc != APR_SUCCESS) {
-            retval = rc;
-        }
-    }
-
     return retval;
 }
 
 int modperl_response_handler(request_rec *r)
 {
     MP_dDCFG;
-    apr_status_t retval;
-
-#ifdef USE_ITHREADS
-    pTHX;
-    modperl_interp_t *interp;
-#endif
+    apr_status_t retval, rc;
+    MP_dINTERP;
 
     if (!strEQ(r->handler, "modperl")) {
         return DECLINED;
     }
 
-#ifdef USE_ITHREADS
-    interp = modperl_interp_select(r, r->connection, r->server);
-    aTHX = interp->perl;
-#endif
+    MP_INTERPa(r, r->connection, r->server);
 
     /* default is -SetupEnv, add if PerlOption +SetupEnv */
     if (MpDirSETUP_ENV(dcfg)) {
         modperl_env_request_populate(aTHX_ r);
     }
 
-    retval = modperl_response_handler_run(r, TRUE);
-
-#ifdef USE_ITHREADS
-    if (MpInterpPUTBACK(interp)) {
-        /* PerlInterpScope handler */
-        modperl_interp_unselect(interp);
+    retval = modperl_response_handler_run(r);
+    rc = modperl_response_finish(r);
+    if (rc != APR_SUCCESS) {
+        retval = rc;
     }
-#endif
+
+    MP_INTERP_PUTBACK(interp, aTHX);
 
     return retval;
 }
@@ -1049,22 +1067,13 @@ int modperl_response_handler_cgi(request_rec *r)
     GV *h_stdin, *h_stdout;
     apr_status_t retval, rc;
     MP_dRCFG;
-#ifdef USE_ITHREADS
-    pTHX;
-    modperl_interp_t *interp;
-#endif
+    MP_dINTERP;
 
     if (!strEQ(r->handler, "perl-script")) {
         return DECLINED;
     }
 
-#ifdef USE_ITHREADS
-    interp = modperl_interp_select(r, r->connection, r->server);
-    aTHX = interp->perl;
-    if (MpInterpPUTBACK(interp)) {
-        rcfg->interp = interp;
-    }
-#endif
+    MP_INTERPa(r, r->connection, r->server);
 
     modperl_perl_global_request_save(aTHX_ r);
 
@@ -1086,7 +1095,7 @@ int modperl_response_handler_cgi(request_rec *r)
 
     modperl_env_request_tie(aTHX_ r);
 
-    retval = modperl_response_handler_run(r, FALSE);
+    retval = modperl_response_handler_run(r);
 
     modperl_env_request_untie(aTHX_ r);
 
@@ -1096,13 +1105,7 @@ int modperl_response_handler_cgi(request_rec *r)
     modperl_io_restore_stdout(aTHX_ h_stdout);
     FREETMPS;LEAVE;
 
-#ifdef USE_ITHREADS
-    if (MpInterpPUTBACK(interp)) {
-        /* PerlInterpScope handler */
-        modperl_interp_unselect(interp);
-        rcfg->interp = NULL;
-    }
-#endif
+    MP_INTERP_PUTBACK(interp, aTHX);
 
     /* flush output buffer after interpreter is putback */
     rc = modperl_response_finish(r);
@@ -1132,7 +1135,7 @@ const void *modperl_suck_in_ugly_hack(void)
 }
 
 module AP_MODULE_DECLARE_DATA perl_module = {
-    STANDARD20_MODULE_STUFF, 
+    STANDARD20_MODULE_STUFF,
     modperl_config_dir_create, /* dir config creater */
     modperl_config_dir_merge,  /* dir merger --- default is to override */
     modperl_config_srv_create, /* server config */
@@ -1140,3 +1143,10 @@ module AP_MODULE_DECLARE_DATA perl_module = {
     modperl_cmds,              /* table of config file commands       */
     modperl_register_hooks,    /* register hooks */
 };
+
+/*
+ * Local Variables:
+ * c-basic-offset: 4
+ * indent-tabs-mode: nil
+ * End:
+ */
