@@ -1,8 +1,9 @@
-/* Copyright 2000-2005 The Apache Software Foundation
- *
- * Licensed under the Apache License, Version 2.0 (the "License");
- * you may not use this file except in compliance with the License.
- * You may obtain a copy of the License at
+/* Licensed to the Apache Software Foundation (ASF) under one or more
+ * contributor license agreements.  See the NOTICE file distributed with
+ * this work for additional information regarding copyright ownership.
+ * The ASF licenses this file to You under the Apache License, Version 2.0
+ * (the "License"); you may not use this file except in compliance with
+ * the License.  You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
@@ -61,7 +62,7 @@ static SV *modperl_hv_request_find(pTHX_ SV *in, char *classname, CV *cv)
 {
     static char *r_keys[] = { "r", "_r", NULL };
     HV *hv = (HV *)SvRV(in);
-    SV *sv = Nullsv;
+    SV *sv = (SV *)NULL;
     int i;
 
     for (i=0; r_keys[i]; i++) {
@@ -116,12 +117,12 @@ MP_INLINE server_rec *modperl_sv2server_rec(pTHX_ SV *sv)
 
 MP_INLINE request_rec *modperl_sv2request_rec(pTHX_ SV *sv)
 {
-    return modperl_xs_sv2request_rec(aTHX_ sv, NULL, Nullcv);
+    return modperl_xs_sv2request_rec(aTHX_ sv, NULL, (CV *)NULL);
 }
 
 request_rec *modperl_xs_sv2request_rec(pTHX_ SV *in, char *classname, CV *cv)
 {
-    SV *sv = Nullsv;
+    SV *sv = (SV *)NULL;
     MAGIC *mg;
 
     if (SvROK(in)) {
@@ -176,7 +177,7 @@ MP_INLINE SV *modperl_newSVsv_obj(pTHX_ SV *stashsv, SV *obj)
 
     if (!obj) {
         obj = stashsv;
-        stashsv = Nullsv;
+        stashsv = (SV *)NULL;
     }
 
     newobj = newSVsv(obj);
@@ -193,7 +194,7 @@ MP_INLINE SV *modperl_ptr2obj(pTHX_ char *classname, void *ptr)
 {
     SV *sv = newSV(0);
 
-    MP_TRACE_h(MP_FUNC, "sv_setref_pv(%s, 0x%lx)\n",
+    MP_TRACE_h(MP_FUNC, "sv_setref_pv(%s, 0x%lx)",
                classname, (unsigned long)ptr);
     sv_setref_pv(sv, classname, ptr);
 
@@ -264,10 +265,10 @@ void **modperl_xs_dl_handles_get(pTHX)
     void **handles;
 
     if (!librefs) {
-	MP_TRACE_r(MP_FUNC,
-                   "Could not get @%s for unloading.\n",
+        MP_TRACE_r(MP_FUNC,
+                   "Could not get @%s for unloading.",
                    dl_librefs);
-	return NULL;
+        return NULL;
     }
 
     if (!(AvFILL(librefs) >= 0)) {
@@ -278,23 +279,23 @@ void **modperl_xs_dl_handles_get(pTHX)
     handles = (void **)malloc(sizeof(void *) * (AvFILL(librefs)+2));
 
     for (i=0; i<=AvFILL(librefs); i++) {
-	void *handle;
-	SV *handle_sv = *av_fetch(librefs, i, FALSE);
-	SV *module_sv = *av_fetch(modules, i, FALSE);
+        void *handle;
+        SV *handle_sv = *av_fetch(librefs, i, FALSE);
+        SV *module_sv = *av_fetch(modules, i, FALSE);
 
-	if(!handle_sv) {
-	    MP_TRACE_r(MP_FUNC,
-                       "Could not fetch $%s[%d]!\n",
+        if(!handle_sv) {
+            MP_TRACE_r(MP_FUNC,
+                       "Could not fetch $%s[%d]!",
                        dl_librefs, (int)i);
-	    continue;
-	}
-	handle = INT2PTR(void *, SvIV(handle_sv));
+            continue;
+        }
+        handle = INT2PTR(void *, SvIV(handle_sv));
 
-	MP_TRACE_r(MP_FUNC, "%s dl handle == 0x%lx\n",
+        MP_TRACE_r(MP_FUNC, "%s dl handle == 0x%lx",
                    SvPVX(module_sv), (unsigned long)handle);
-	if (handle) {
-	    handles[i] = handle;
-	}
+        if (handle) {
+            handles[i] = handle;
+        }
     }
 
     av_clear(modules);
@@ -310,11 +311,11 @@ void modperl_xs_dl_handles_close(void **handles)
     int i;
 
     if (!handles) {
-	return;
+        return;
     }
 
     for (i=0; handles[i]; i++) {
-        MP_TRACE_r(MP_FUNC, "close 0x%lx\n", (unsigned long)handles[i]);
+        MP_TRACE_r(MP_FUNC, "close 0x%lx", (unsigned long)handles[i]);
         modperl_sys_dlclose(handles[i]);
     }
 
@@ -343,7 +344,7 @@ static void modperl_av_remove_entry(pTHX_ AV *av, I32 index)
     sv_free((SV *)tmpav);
 }
 
-static void modperl_package_unload_dynamic(pTHX_ const char *package, 
+static void modperl_package_unload_dynamic(pTHX_ const char *package,
                                            I32 dl_index)
 {
     AV *librefs = get_av(dl_librefs, 0);
@@ -355,7 +356,7 @@ static void modperl_package_unload_dynamic(pTHX_ const char *package,
     modperl_av_remove_entry(aTHX_ get_av(dl_librefs, 0), dl_index);
     modperl_av_remove_entry(aTHX_ get_av(dl_modules, 0), dl_index);
 
-    return;    
+    return;
 }
 
 static int modperl_package_is_dynamic(pTHX_ const char *package,
@@ -414,12 +415,12 @@ HE *modperl_perl_hv_fetch_he(pTHX_ HV *hv,
 
 #ifdef HvREHASH
     if (HvREHASH(hv)) {
-	PERL_HASH_INTERNAL(hash, key, klen);
+        PERL_HASH_INTERNAL(hash, key, klen);
     }
     else
 #endif
     if (!hash) {
-	PERL_HASH(hash, key, klen);
+        PERL_HASH(hash, key, klen);
     }
 
     entry = ((HE**)HvARRAY(hv))[hash & (I32)xhv->xhv_max];
@@ -448,7 +449,7 @@ void modperl_str_toupper(char *str)
     }
 }
 
-/* XXX: same as Perl_do_sprintf(); 
+/* XXX: same as Perl_do_sprintf();
  * but Perl_do_sprintf() is not part of the "public" api
  */
 void modperl_perl_do_sprintf(pTHX_ SV *sv, I32 len, SV **sarg)
@@ -457,7 +458,7 @@ void modperl_perl_do_sprintf(pTHX_ SV *sv, I32 len, SV **sarg)
     char *pat = SvPV(*sarg, patlen);
     bool do_taint = FALSE;
 
-    sv_vsetpvfn(sv, pat, patlen, Null(va_list*), sarg + 1, len - 1, &do_taint);
+    sv_vsetpvfn(sv, pat, patlen, (va_list *)NULL, sarg + 1, len - 1, &do_taint);
     SvSETMAGIC(sv);
     if (do_taint) {
         SvTAINTED_on(sv);
@@ -469,26 +470,25 @@ void modperl_perl_call_list(pTHX_ AV *subs, const char *name)
     I32 i, oldscope = PL_scopestack_ix;
     SV **ary = AvARRAY(subs);
 
-    MP_TRACE_g(MP_FUNC, "pid %lu" MP_TRACEf_TID MP_TRACEf_PERLID
-               " running %d %s subs",
-               (unsigned long)getpid(), MP_TRACEv_TID_ MP_TRACEv_PERLID_
+    MP_TRACE_g(MP_FUNC, MP_TRACEf_PERLID
+               " running %d %s subs", MP_TRACEv_PERLID_
                AvFILLp(subs)+1, name);
 
     for (i=0; i<=AvFILLp(subs); i++) {
-	CV *cv = (CV*)ary[i];
-	SV *atsv = ERRSV;
+        CV *cv = (CV*)ary[i];
+        SV *atsv = ERRSV;
 
-	PUSHMARK(PL_stack_sp);
-	call_sv((SV*)cv, G_EVAL|G_DISCARD);
+        PUSHMARK(PL_stack_sp);
+        call_sv((SV*)cv, G_EVAL|G_DISCARD);
 
-	if (SvCUR(atsv)) {
+        if (SvCUR(atsv)) {
             Perl_sv_catpvf(aTHX_ atsv, "%s failed--call queue aborted",
                            name);
-	    while (PL_scopestack_ix > oldscope) {
-		LEAVE;
+            while (PL_scopestack_ix > oldscope) {
+                LEAVE;
             }
             Perl_croak(aTHX_ "%s", SvPVX(atsv));
-	}
+        }
     }
 }
 
@@ -496,7 +496,7 @@ void modperl_perl_exit(pTHX_ int status)
 {
     ENTER;
     SAVESPTR(PL_diehook);
-    PL_diehook = Nullsv; 
+    PL_diehook = (SV *)NULL;
     modperl_croak(aTHX_ MODPERL_RC_EXIT, "ModPerl::Util::exit");
 }
 
@@ -505,7 +505,7 @@ MP_INLINE SV *modperl_dir_config(pTHX_ request_rec *r, server_rec *s,
 {
     SV *retval = &PL_sv_undef;
 
-    if (r && r->per_dir_config) {				   
+    if (r && r->per_dir_config) {
         MP_dDCFG;
         retval = modperl_table_get_set(aTHX_ dcfg->configvars,
                                        key, sv_val, FALSE);
@@ -531,31 +531,31 @@ SV *modperl_table_get_set(pTHX_ apr_table_t *table, char *key,
 {
     SV *retval = &PL_sv_undef;
 
-    if (table == NULL) { 
+    if (table == NULL) {
         /* do nothing */
     }
-    else if (key == NULL) { 
+    else if (key == NULL) {
         retval = modperl_hash_tie(aTHX_ "APR::Table",
-                                  Nullsv, (void*)table); 
+                                  (SV *)NULL, (void*)table);
     }
     else if (!sv_val) { /* no val was passed */
-        char *val; 
-        if ((val = (char *)apr_table_get(table, key))) { 
-            retval = newSVpv(val, 0); 
-        } 
-        else { 
-            retval = newSV(0); 
-        } 
-        if (do_taint) { 
-            SvTAINTED_on(retval); 
-        } 
+        char *val;
+        if ((val = (char *)apr_table_get(table, key))) {
+            retval = newSVpv(val, 0);
+        }
+        else {
+            retval = newSV(0);
+        }
+        if (do_taint) {
+            SvTAINTED_on(retval);
+        }
     }
     else if (!SvOK(sv_val)) { /* val was passed in as undef */
-        apr_table_unset(table, key); 
+        apr_table_unset(table, key);
     }
-    else { 
+    else {
         apr_table_set(table, key, SvPV_nolen(sv_val));
-    } 
+    }
 
     return retval;
 }
@@ -628,10 +628,10 @@ MP_INLINE SV *modperl_slurp_filename(pTHX_ request_rec *r, int tainted)
     rc = apr_file_read(file, SvPVX(sv), &size);
     SLURP_SUCCESS("reading");
 
-    MP_TRACE_o(MP_FUNC, "read %d bytes from '%s'\n", size, r->filename);
+    MP_TRACE_o(MP_FUNC, "read %d bytes from '%s'", size, r->filename);
 
     if (r->finfo.size != size) {
-        SvREFCNT_dec(sv); 
+        SvREFCNT_dec(sv);
         Perl_croak(aTHX_ "Error: read %d bytes, expected %d ('%s')",
                    size, (apr_size_t)r->finfo.size, r->filename);
     }
@@ -705,10 +705,10 @@ char *modperl_file2package(apr_pool_t *p, const char *file)
 
 SV *modperl_apr_array_header2avrv(pTHX_ apr_array_header_t *array)
 {
-    AV *av = newAV(); 
+    AV *av = newAV();
 
     if (array) {
-        int i; 
+        int i;
         for (i = 0; i < array->nelts; i++) {
             av_push(av, newSVpv(((char **)array->elts)[i], 0));
         }
@@ -734,25 +734,25 @@ apr_array_header_t *modperl_avrv2apr_array_header(pTHX_ apr_pool_t *p,
     for (i = 0; i <= av_size; i++) {
         SV *sv = *av_fetch(av, i, FALSE);
         char **entry = (char **)apr_array_push(array);
-        *entry = apr_pstrdup(p, SvPV(sv, PL_na));
+        *entry = apr_pstrdup(p, SvPV_nolen(sv));
     }
 
     return array;
 }
 
 /* Remove a package from %INC */
-static void modperl_package_delete_from_inc(pTHX_ const char *package)  
+static void modperl_package_delete_from_inc(pTHX_ const char *package)
 {
     int len;
     char *filename = package2filename(package, &len);
-    hv_delete(GvHVn(PL_incgv), filename, len, G_DISCARD);
+    (void)hv_delete(GvHVn(PL_incgv), filename, len, G_DISCARD);
     free(filename);
 }
 
 /* Destroy a package's stash */
 #define MP_STASH_SUBSTASH(key, len) ((len >= 2) &&                  \
                                      (key[len-1] == ':') &&         \
-                                     (key[len-2] == ':'))   
+                                     (key[len-2] == ':'))
 #define MP_STASH_DEBUGGER(key, len) ((len >= 2) &&                  \
                                      (key[0] == '_') &&             \
                                      (key[1] == '<'))
@@ -775,7 +775,7 @@ static void modperl_package_clear_stash(pTHX_ const char *package)
                  * unload
                  */
                 if (GvSTASH(val) == stash) {
-                    hv_delete(stash, key, len, G_DISCARD);
+                    (void)hv_delete(stash, key, len, G_DISCARD);
                 }
             }
         }
@@ -818,7 +818,7 @@ void modperl_restart_count_inc(server_rec *base_server)
         *counter = 1;
         apr_pool_userdata_set(counter, MP_RESTART_COUNT_KEY,
                               apr_pool_cleanup_null, p);
-    }    
+    }
 }
 
 int modperl_restart_count(void)
@@ -828,3 +828,441 @@ int modperl_restart_count(void)
                           modperl_global_get_server_rec()->process->pool);
     return data ? *(int *)data : 0;
  }
+
+static MP_INLINE
+apr_status_t modperl_cleanup_pnotes(void *data) {
+    modperl_pnotes_t *pnotes = data;
+
+    dTHXa(pnotes->interp->perl);
+    MP_ASSERT_CONTEXT(aTHX);
+
+    SvREFCNT_dec(pnotes->pnotes);
+    pnotes->pnotes = NULL;
+    pnotes->pool = NULL;
+
+    MP_INTERP_PUTBACK(pnotes->interp, aTHX);
+    return APR_SUCCESS;
+}
+
+void modperl_pnotes_kill(void *data) {
+    modperl_pnotes_t *pnotes = data;
+
+    if( !pnotes->pnotes ) return;
+
+    apr_pool_cleanup_kill(pnotes->pool, pnotes, modperl_cleanup_pnotes);
+    modperl_cleanup_pnotes(pnotes);
+}
+
+SV *modperl_pnotes(pTHX_ modperl_pnotes_t *pnotes, SV *key, SV *val,
+                   apr_pool_t *pool) {
+    SV *retval = (SV *)NULL;
+
+    if (!pnotes->pnotes) {
+        pnotes->pool = pool;
+#ifdef USE_ITHREADS
+        pnotes->interp = modperl_thx_interp_get(aTHX);
+        pnotes->interp->refcnt++;
+        MP_TRACE_i(MP_FUNC, "TO: (0x%lx)->refcnt incremented to %ld",
+                   pnotes->interp, pnotes->interp->refcnt);
+#endif
+        pnotes->pnotes = newHV();
+        apr_pool_cleanup_register(pool, pnotes,
+                                  modperl_cleanup_pnotes,
+                                  apr_pool_cleanup_null);
+    }
+
+    if (key) {
+        STRLEN len;
+        char *k = SvPV(key, len);
+
+        if (val) {
+            retval = *hv_store(pnotes->pnotes, k, len, SvREFCNT_inc(val), 0);
+        }
+        else if (hv_exists(pnotes->pnotes, k, len)) {
+            retval = *hv_fetch(pnotes->pnotes, k, len, FALSE);
+        }
+
+        return retval ? SvREFCNT_inc(retval) : &PL_sv_undef;
+    }
+    return newRV_inc((SV *)pnotes->pnotes);
+}
+
+U16 *modperl_code_attrs(pTHX_ CV *cv) {
+    MAGIC *mg;    
+
+    if (!(SvMAGICAL(cv) && (mg = mg_find((SV*)cv, PERL_MAGIC_ext)))) {
+       sv_magic((SV*)cv, (SV *)NULL, PERL_MAGIC_ext, NULL, -1); 
+    }
+
+    mg = mg_find((SV*)cv, PERL_MAGIC_ext);
+    return &(mg->mg_private);
+}
+
+#if AP_SERVER_MAJORVERSION_NUMBER>2 || \
+    (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER>=3)
+
+static apr_hash_t *global_authz_providers = NULL;
+static apr_hash_t *global_authn_providers = NULL;
+
+typedef struct {
+    SV *cb1;
+    SV *cb2;
+    modperl_handler_t *cb1_handler;
+    modperl_handler_t *cb2_handler;
+} auth_callback;
+
+static apr_status_t cleanup_perl_global_providers(void *ctx)
+{
+    global_authz_providers = NULL;
+    global_authn_providers = NULL;
+    return APR_SUCCESS;
+}
+
+static authz_status perl_check_authorization(request_rec *r,
+                                             const char *require_args,
+                                             const void *parsed_require_args)
+{
+    authz_status ret = AUTHZ_DENIED;
+    int count;
+    AV *args = Nullav;
+    const char *key;
+    auth_callback *ab;
+    MP_dINTERPa(r, NULL, NULL);
+
+    if (global_authz_providers == NULL) {
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    key = apr_table_get(r->notes, AUTHZ_PROVIDER_NAME_NOTE);
+    ab = apr_hash_get(global_authz_providers, key, APR_HASH_KEY_STRING);
+    if (ab == NULL) {
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    if (ab->cb1 == NULL) {
+        if (ab->cb1_handler == NULL) {
+            MP_INTERP_PUTBACK(interp, aTHX);
+            return ret;
+        }
+
+        modperl_handler_make_args(aTHX_ &args, "Apache2::RequestRec", r,
+                                  "PV", require_args, NULL);
+        ret = modperl_callback(aTHX_ ab->cb1_handler, r->pool, r, r->server,
+                               args);
+        SvREFCNT_dec((SV*)args);
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    {
+        dSP;
+        ENTER;
+        SAVETMPS;
+        PUSHMARK(SP);
+        XPUSHs(sv_2mortal(modperl_ptr2obj(aTHX_ "Apache2::RequestRec", r)));
+        XPUSHs(sv_2mortal(newSVpv(require_args, 0)));
+        PUTBACK;
+        count = call_sv(ab->cb1, G_SCALAR);
+        SPAGAIN;
+
+        if (count == 1) {
+            ret = (authz_status) POPi;
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+    }
+
+    MP_INTERP_PUTBACK(interp, aTHX);
+    return ret;
+}
+
+static const char *perl_parse_require_line(cmd_parms *cmd,
+                                           const char *require_line,
+                                           const void **parsed_require_line)
+{
+    char *ret = NULL;
+    void *key;
+    auth_callback *ab;
+
+    if (global_authz_providers == NULL ||
+        apr_hash_count(global_authz_providers) == 0)
+    {
+        return NULL;
+    }
+
+    apr_pool_userdata_get(&key, AUTHZ_PROVIDER_NAME_NOTE, cmd->temp_pool);
+    ab = apr_hash_get(global_authz_providers, (char *) key, APR_HASH_KEY_STRING);
+    if (ab == NULL || ab->cb2 == NULL) {
+        return NULL;
+    }
+
+    {
+        /* PerlAddAuthzProvider currently does not support an optional second
+         * handler, so ab->cb2 should always be NULL above and we will never get
+         * here. If such support is added in the future then this code will be
+         * reached, but cannot succeed in the absence of an interpreter. The
+         * second handler would be called at init to check a Require line for
+         * errors, but in the current design there is no interpreter available
+         * at that time.
+         */
+        MP_dINTERP_POOLa(cmd->pool, cmd->server);
+        if (!MP_HAS_INTERP(interp)) {
+	    return "Require handler is not currently supported in this context";
+	}
+
+        {
+            SV *ret_sv;
+            int count;
+            dSP;
+
+            ENTER;
+            SAVETMPS;
+            PUSHMARK(SP);
+            XPUSHs(sv_2mortal(modperl_ptr2obj(aTHX_ "Apache2::CmdParms", cmd)));
+            XPUSHs(sv_2mortal(newSVpv(require_line, 0)));
+            PUTBACK;
+            count = call_sv(ab->cb2, G_SCALAR);
+            SPAGAIN;
+
+            if (count == 1) {
+                ret_sv = POPs;
+                if (SvOK(ret_sv)) {
+                    char *tmp = SvPV_nolen(ret_sv);
+                    if (*tmp != '\0') {
+                        ret = apr_pstrdup(cmd->pool, tmp);
+                    }
+                }
+            }
+
+            PUTBACK;
+            FREETMPS;
+            LEAVE;
+        }
+
+        MP_INTERP_PUTBACK(interp, aTHX);
+    }
+    return ret;
+}
+
+static authn_status perl_check_password(request_rec *r, const char *user,
+                                        const char *password)
+{
+    authn_status ret = AUTH_DENIED;
+    int count;
+    AV *args = Nullav;
+    const char *key;
+    auth_callback *ab;
+    MP_dINTERPa(r, NULL, NULL);
+
+    if (global_authn_providers == NULL) {
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    key = apr_table_get(r->notes, AUTHN_PROVIDER_NAME_NOTE);
+    ab = apr_hash_get(global_authn_providers, key,
+                                     APR_HASH_KEY_STRING);
+    if (ab == NULL || ab->cb1) {
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    if (ab->cb1 == NULL) {
+        if (ab->cb1_handler == NULL) {
+            MP_INTERP_PUTBACK(interp, aTHX);
+            return ret;
+        }
+
+        modperl_handler_make_args(aTHX_ &args, "Apache2::RequestRec", r,
+                                  "PV", user,
+                                  "PV", password, NULL);
+        ret = modperl_callback(aTHX_ ab->cb1_handler, r->pool, r, r->server,
+                               args);
+        SvREFCNT_dec((SV*)args);
+        MP_INTERP_PUTBACK(interp, aTHX);
+        return ret;
+    }
+
+    {
+        dSP;
+        ENTER;
+        SAVETMPS;
+        PUSHMARK(SP);
+        XPUSHs(sv_2mortal(modperl_ptr2obj(aTHX_ "Apache2::RequestRec", r)));
+        XPUSHs(sv_2mortal(newSVpv(user, 0)));
+        XPUSHs(sv_2mortal(newSVpv(password, 0)));
+        PUTBACK;
+        count = call_sv(ab->cb1, G_SCALAR);
+        SPAGAIN;
+
+        if (count == 1) {
+            ret = (authn_status) POPi;
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+    }
+
+    MP_INTERP_PUTBACK(interp, aTHX);
+    return ret;
+}
+
+static authn_status perl_get_realm_hash(request_rec *r, const char *user,
+                                        const char *realm, char **rethash)
+{
+    authn_status ret = AUTH_USER_NOT_FOUND;
+    const char *key;
+    auth_callback *ab;
+
+    if (global_authn_providers == NULL ||
+        apr_hash_count(global_authn_providers) == 0)
+    {
+        return AUTH_GENERAL_ERROR;
+    }
+
+    key = apr_table_get(r->notes, AUTHN_PROVIDER_NAME_NOTE);
+    ab = apr_hash_get(global_authn_providers, key, APR_HASH_KEY_STRING);
+    if (ab == NULL || ab->cb2 == NULL) {
+        return AUTH_GENERAL_ERROR;
+    }
+
+    {
+        /* PerlAddAuthnProvider currently does not support an optional second
+         * handler, so ab->cb2 should always be NULL above and we will never get
+         * here. If such support is added in the future then this code will be
+         * reached. Unlike the PerlAddAuthzProvider case, the second handler here
+         * would be called during request_rec processing to obtain a password hash
+         * for the realm so there should be no problem grabbing an interpreter.
+         */
+        MP_dINTERPa(r, NULL, NULL);
+
+        {
+            SV* rh = sv_2mortal(newSVpv("", 0));
+            int count;
+            dSP;
+
+            ENTER;
+            SAVETMPS;
+            PUSHMARK(SP);
+            XPUSHs(sv_2mortal(modperl_ptr2obj(aTHX_ "Apache2::RequestRec", r)));
+            XPUSHs(sv_2mortal(newSVpv(user, 0)));
+            XPUSHs(sv_2mortal(newSVpv(realm, 0)));
+            XPUSHs(newRV_noinc(rh));
+            PUTBACK;
+            count = call_sv(ab->cb2, G_SCALAR);
+            SPAGAIN;
+
+            if (count == 1) {
+                const char *tmp = SvPV_nolen(rh);
+                ret = (authn_status) POPi;
+                if (*tmp != '\0') {
+                    *rethash = apr_pstrdup(r->pool, tmp);
+                }
+            }
+
+            PUTBACK;
+            FREETMPS;
+            LEAVE;
+        }
+
+        MP_INTERP_PUTBACK(interp, aTHX);
+    }
+
+    return ret;
+}
+
+static const authz_provider authz_perl_provider = { perl_check_authorization,
+                                                    perl_parse_require_line };
+
+static const authn_provider authn_perl_provider = { perl_check_password,
+                                                    perl_get_realm_hash };
+
+static apr_status_t register_auth_provider(apr_pool_t *pool,
+                                           const char *provider_group,
+                                           const char *provider_name,
+                                           const char *provider_version,
+                                           auth_callback *ab, int type)
+{
+    void *provider_ = NULL;
+
+    if (global_authz_providers == NULL) {
+        global_authz_providers = apr_hash_make(pool);
+        global_authn_providers = apr_hash_make(pool);
+        /* We have to use pre_cleanup here, otherwise this cleanup method
+         * would be called after another cleanup method which unloads
+         * mod_perl module.
+         */
+        apr_pool_pre_cleanup_register(pool, NULL,
+                                      cleanup_perl_global_providers);
+    }
+
+    if (strcmp(provider_group, AUTHZ_PROVIDER_GROUP) == 0) {
+        provider_ = (void *) &authz_perl_provider;
+        apr_hash_set(global_authz_providers, provider_name,
+                     APR_HASH_KEY_STRING, ab);
+    }
+    else {
+        provider_ = (void *) &authn_perl_provider;
+        apr_hash_set(global_authn_providers, provider_name,
+                     APR_HASH_KEY_STRING, ab);
+    }
+
+    return ap_register_auth_provider(pool, provider_group, provider_name,
+                                     provider_version, provider_, type);
+
+}
+
+apr_status_t modperl_register_auth_provider(apr_pool_t *pool,
+                                            const char *provider_group,
+                                            const char *provider_name,
+                                            const char *provider_version,
+                                            SV *callback1, SV *callback2,
+                                            int type)
+{
+    char *provider_name_dup;
+    auth_callback *ab = NULL;
+
+    provider_name_dup = apr_pstrdup(pool, provider_name);
+    ab = apr_pcalloc(pool, sizeof(auth_callback));
+    ab->cb1 = callback1;
+    ab->cb2 = callback2;
+
+    return register_auth_provider(pool, provider_group, provider_name_dup,
+                                  provider_version, ab, type);
+}
+
+apr_status_t modperl_register_auth_provider_name(apr_pool_t *pool,
+                                                 const char *provider_group,
+                                                 const char *provider_name,
+                                                 const char *provider_version,
+                                                 const char *callback1,
+                                                 const char *callback2,
+                                                 int type)
+{
+    char *provider_name_dup;
+    auth_callback *ab = NULL;
+
+    provider_name_dup = apr_pstrdup(pool, provider_name);
+    ab = apr_pcalloc(pool, sizeof(auth_callback));
+    ab->cb1_handler = modperl_handler_new(pool, callback1);
+    if (callback2) {
+        ab->cb2_handler = modperl_handler_new(pool, callback2);
+    }
+
+    return register_auth_provider(pool, provider_group, provider_name_dup,
+                                  provider_version, ab, type);
+}
+
+#endif /* httpd-2.4 */
+
+/*
+ * Local Variables:
+ * c-basic-offset: 4
+ * indent-tabs-mode: nil
+ * End:
+ */
